//! VTableLookupDataHandler data class for mscw and energy reconstruction

#ifndef VTableLookupDataHandler_H
#define VTableLookupDataHandler_H

#include "VEmissionHeightCalculator.h"
#include "VDeadTime.h"
#include "VEffectiveAreaCalculatorMCHistograms.h"
#include "VMonteCarloRunHeader.h"
#include "VTableLookupRunParameter.h"
#include "VUtilities.h"

#include "TChain.h"
#include "TChainElement.h"
#include "TDirectory.h"
#include "TError.h"
#include "TFile.h"
#include "TH1D.h"
#include "TH2D.h"
#include "TList.h"
#include "TMath.h"
#include "TObjArray.h"
#include "TRandom3.h"
#include "TKey.h"

#include "Cshowerpars.h"
#include "Cmodel3Dpars.h"
#include "Ctelconfig.h"
#include "Ctpars.h"

#include <bitset>
#include <cmath>
#include <iostream>
#include <map>
#include <string>
#include <vector>

using namespace std;

class VTableLookupDataHandler
{
	private:
		unsigned int fDebug;
		
		VTableLookupRunParameter* fTLRunParameter;  //! lookup parameters
		
		UInt_t fNMethods;                            //!< number of direction and core reconstruction methods in evndisp
		vector< string > finputfile;              //!< input file name
		string foutputfile;                       //!< output file name
		TFile* fOutFile;                          //!< point to output file
		bool   fwrite;                            //!< true for table filling
		
		unsigned int fNTel;                       //!< number of telescopes
		unsigned int fNTelComb;                   //!< number of telescope combinations
		int fNEntries;                            //!< total number of events in input tree
		int fEventCounter;                        //!< event counter for input tree
		int fMethod;                              //!< which direction and core reconstruction method data should be used
		
		double fMaxTotalTime;                     //!< time to analyse in a run (in [s])
		double fTotalTime;                        //!< time passed in analysing the run (in [s])
		double fTotalTime0;                       //!< time of first event (in [s])
		
		VEmissionHeightCalculator* fEmissionHeightCalculator;
		
		double fSelectRandom;
		int fSelectRandomSeed;
		TRandom3* fRandom;
		
		// MC parameter
		unsigned int fMinImages;
		double fMC_distance_to_cameracenter_min;
		double fMC_distance_to_cameracenter_max;
		
		double fMCSpectralIndex;
		double fMCMinEnergy;
		double fMCMaxEnergy;
		double fSpectralIndex;
		
		vector< double > fNoiseLevel;
		vector< double > fCurrentNoiseLevel;
		
		// input trees
		int fEventDisplayFileFormat;
		TChain* fTshowerpars;
		TChain* fTshowerpars_QCCut;
		Cshowerpars* fshowerpars;
		TChain* fTmodel3Dpars;
		Cmodel3Dpars* fmodel3Dpars;
		TKey* fKeyModel3D;
		TChain* fTtelconfig;
		Ctelconfig* ftelconfig;
		vector< TChain* > fTtpars;
		vector< Ctpars* > ftpars;
                TChain *fFrogspars;
		
		double fEventWeight;
		
		// MC energy histograms
		TH1D* hE0mc;
		TH2D* hXYmc;
		TH2D* hDE0mc;
		TH2D* hWE0mc;
		TH1D* hZe;
		TH1D* hE0trig;
		TH2D* hXYtrig;
		TH2D* hDE0trig;
		TH2D* hWE0trig;
		TList* hisList;
		
		// telescope positions from fTtelconfig
		double fTelX[VDST_MAXTELESCOPES];
		double fTelY[VDST_MAXTELESCOPES];
		double fTelZ[VDST_MAXTELESCOPES];
		double fFocalLength[VDST_MAXTELESCOPES];
		ULong64_t fTel_type[VDST_MAXTELESCOPES];
		map<ULong64_t, unsigned int > fList_of_Tel_type;
		map<ULong64_t, unsigned int >::iterator fList_of_Tel_type_iterator;
		vector< unsigned int > fTel_type_counter;
		// telescope pointing
                float  fArrayElevation;
                float  fArrayAzimuth;
		double fTelElevation[VDST_MAXTELESCOPES];
		double fTelAzimuth[VDST_MAXTELESCOPES];
		double fWobbleN;
		double fWobbleE;
                float  fArrayPointing_Elevation;
                float  fArrayPointing_Azimuth;
		
		// output trees
		TTree* fOTree;
		bool fShortTree;                          //!< use short version of output tree
		bool bWriteMCPars;
		
		// cut statistics
		bool fEventStatus;
		unsigned int fNStats_All;
		unsigned int fNStats_Rec;
		unsigned int fNStats_NImagesCut;
		unsigned int fNStats_Chi2Cut;
		unsigned int fNStats_CoreErrorCut;
		unsigned int fNStats_WobbleCut;
		unsigned int fNStats_WobbleMinCut;
		unsigned int fNStats_WobbleMaxCut;
		
		void   calcDistances( int );                //!< calculate distances between telescopes and shower core
		void   calcEmissionHeights();
		double calculateMeanNoiseLevel( bool bCurrentNoiseLevel = false );
                bool   checkCoreReconstructionQuality( unsigned int i );
		bool   checkIfFilesInChainAreRecovered( TChain* c );
		void   copyMCHistograms();
		bool   copyMCRunheader();
		void   copyMCTree();
		void   copy_telconfig();
		void   initializeTelTypeVector();
		int    fillNextEvent( bool bShort );
		void   printCutStatistics();
		bool   randomSelected();
		void   resetImageParameters();
		void   resetImageParameters( unsigned int i );
		void   setEventWeightfromMCSpectrum();
		void   setSelectRandom( double iX, int iS );
		void   writeDeadTimeHistograms();
		
	public:
	
		//  data written to output file
		
		int runNumber;
		int eventNumber;
		int MJD;
		double time;
		
		bool fIsMC;                               //!< data is MC
		int    fMCPrimary;
		double fMCEnergy;                         //!< MC energy
		double fMCxcore;
		double fMCycore;
		double fMCxcore_SC;
		double fMCycore_SC;
		double fMCxcos;
		double fMCycos;
		double fMCaz;
		double fMCze;
		double fMCxoff;
		double fMCyoff;
		Int_t	       fMCCorsikaRunID;
		Int_t	       fMCCorsikaShowerID;
		Float_t        fMCFirstInteractionHeight;
		Float_t	       fMCFirstInteractionDepth;

		unsigned int LTrigS;
		ULong64_t LTrig;
		unsigned int fNTrig;
		int fNImages;
                UShort_t fNImages_QCTree[VDST_MAXTELESCOPES];
		unsigned int fImgSelS;
		ULong64_t fImgSel;
		bool fImgSel_list[VDST_MAXTELESCOPES];
		unsigned int fImgSel_list_short[VDST_MAXTELESCOPES];
		int          fNTelTypes;
                ULong64_t    fTtypeID[VDST_MAXTELESCOPES];
		unsigned int NImages_Ttype[VDST_MAXTELESCOPES];
		double fimg2_ang;
		double fZe;                               //!< zenith angle
		double fAz;
		double fRA;
		double fDec;
		double fXoff;
		double fYoff;
		double fXoff_derot;
		double fYoff_derot;
		double fstdS;
		double ftheta2;
		double fXcore;
		double fYcore;
		double fXcore_SC;
		double fYcore_SC;
		double fstdP;
		double fchi2;                             //!< chi2 from array reconstruction
                Float_t fchi2_QCTree[VDST_MAXTELESCOPES];
		//
		// {1}
		double fMCEnergyArray [VDST_MAXTELESCOPES];
		float  fmeanPedvar_ImageT[VDST_MAXTELESCOPES];
		float  fmeanPedvar_Image;
		double fdist     [VDST_MAXTELESCOPES];
		double fdist_telType[VDST_MAXTELESCOPES];
		double fsize     [VDST_MAXTELESCOPES];
		double fsize2    [VDST_MAXTELESCOPES];
		double fsizeCorr [VDST_MAXTELESCOPES];
		double fsize_telType[VDST_MAXTELESCOPES];
		double floss     [VDST_MAXTELESCOPES];
		double fLoss_telType[VDST_MAXTELESCOPES];
		double ffracLow  [VDST_MAXTELESCOPES];
		double fmax1     [VDST_MAXTELESCOPES];
		double fmax2     [VDST_MAXTELESCOPES];
		double fmax3     [VDST_MAXTELESCOPES];
		int    fmaxindex1     [VDST_MAXTELESCOPES];
		int    fmaxindex2     [VDST_MAXTELESCOPES];
		int    fmaxindex3     [VDST_MAXTELESCOPES];
		double fwidth    [VDST_MAXTELESCOPES];
		double fwidth_telType[VDST_MAXTELESCOPES];
		double flength   [VDST_MAXTELESCOPES];
		double flength_telType[VDST_MAXTELESCOPES];
		int    fntubes   [VDST_MAXTELESCOPES];
		unsigned short int fnsat[VDST_MAXTELESCOPES];
		unsigned short int fnlowgain[VDST_MAXTELESCOPES];
		double falpha    [VDST_MAXTELESCOPES];
		double flos      [VDST_MAXTELESCOPES];
		double fasym     [VDST_MAXTELESCOPES];
		double fcen_x    [VDST_MAXTELESCOPES];
		double fcen_y    [VDST_MAXTELESCOPES];
		double fcosphi   [VDST_MAXTELESCOPES];
		double fsinphi   [VDST_MAXTELESCOPES];
		double ftgrad_x  [VDST_MAXTELESCOPES];
                double ftgrad_x_telType[VDST_MAXTELESCOPES];
		double ftchisq_x [VDST_MAXTELESCOPES];
		double fweight   [VDST_MAXTELESCOPES];    //!< always 1.
		int    fFitstat  [VDST_MAXTELESCOPES];
		double fR        [VDST_MAXTELESCOPES];    //!< distance from each telescope to reconstructed shower core
		double fR_telType[VDST_MAXTELESCOPES];    //!< distance from each telescope to reconstructed shower core (depending on tel type)
                double fR_short[VDST_MAXTELESCOPES];    //!< distance from each telescope to reconstructed shower core (depending on imgsel)
		double fR_MC      [VDST_MAXTELESCOPES];    //!< distance from each telescope to MC shower core
		double fR_telType_MC[VDST_MAXTELESCOPES];    //!< distance from each telescope to MC shower core (depending on tel type)
                double fR_short_MC[VDST_MAXTELESCOPES];    //!< distance from each telescope to MC shower core (depending on imgsel)
                double ffrogs_goodness_telType[VDST_MAXTELESCOPES];
                vector<float>* ffrogs_goodness_vector;

		
                ///////////////////////////////////////////////////////////////////////////////////////////////////////////
                // results
                ///////////////////////////////////////////////////////////////////////////////////////////////////////////
                
		int    fnmscw;                            //!< number of images used for mscw/mscl/energy calculation
                // MSCW
		double fmscw;                             //!< mean scaled width
		double ftmscw    [VDST_MAXTELESCOPES];    //!< mscw assigned to each telescope
		float  ftmscw_sigma[VDST_MAXTELESCOPES];  //!< mscw  sigma  assigned to each telescope
                // MWR
		float  fmwr;                              //!< mean width ratio

                // MSCL
		double fmscl;                             //!< mean scaled length
		double ftmscl    [VDST_MAXTELESCOPES];    //!< mscl assigned to each telescope
		float  ftmscl_sigma[VDST_MAXTELESCOPES];  //!< mscl  sigma  assigned to each telescope
                // MWL
		float  fmlr;                              //!< mean length ratio

                // reconstructed energy
		double fenergyS;                          //!< reconstructed primary energy
		double fechi2S;                           //!< chi2 from reconstructed primary energy
		double fdES;                              //!< dE from reconstructed primary energy
		double fES       [VDST_MAXTELESCOPES];    //!< energy assigned to each telescope (method 1)
		double fES_short [VDST_MAXTELESCOPES];    //!< energy assigned to each telescope (method 1) (depending on imgsel)

                // time gradient
                double fmsct;                             //!< mean scaled time gradient
                double ftmsct    [VDST_MAXTELESCOPES];    //!< msct assigned to each telescope
                float  ftmsct_sigma[VDST_MAXTELESCOPES];  //!< msct sigma assigned to each telescope
                
                // FROGS goodness
                double fmsc_frgo;                             //!< mean scaled frogs goodness
                double ftmsc_frgo    [VDST_MAXTELESCOPES];    //!< msc_frgo assigned to each telescope
                float  ftmsc_frgo_sigma[VDST_MAXTELESCOPES];  //!< msc_frgo sigma assigned to each telescope
		
		// emission height
		unsigned int fNTelPairs;
		float  fEmissionHeightMean;
		float  fEmissionHeightChi2;
		//!< (note that maximum array length should be larger than MaxNbrTel
		float  fEmissionHeightT[VDST_MAXTELESCOPES];
		
		double fSizeSecondMax;
		double ftheta2_All[25];

                // difference in disp event direction between telescopes
		double fDispDiff;
                
		// Model3D parameters 
		bool fUseModel3DStereoParameters; // overwrite geometric Xoff,Yoff,Xcore,Ycore

		bool fIsModel3D;  // data contains Model3D parameters
		double fSmax3D;   // height of shower maximum (along the shower axis)
		double fsigmaL3D; // longitudinal (3D-length)
		double fsigmaT3D; // transverse (3D-width)
		double fNc3D;     // total number of Cherenkov photons emitted by the shower
		double fXcore3D;  // core location
		double fYcore3D;  // core location
		double fXoff3D;  // model sky direction
		double fYoff3D;  // model sky direction
		double fXoffDeRot3D;  // model sky direction (derotated)
		double fYoffDeRot3D;  // model sky direction (derotated)
		double fGoodness3D;   // model goodness of fit
		double fOmega3D;      // model and hillas: direction difference
		double fDepth3D;      // model: slant depth of shower maximum
		double fRWidth3D;     // model: reduced 3D-width
		double fErrRWidth3D;  // model: error in reduced 3D-width
		double fErrorsigmaT3D;// model: error in 3D-width
		bool fConverged3D;   // model: fit converged
		
		/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		VTableLookupDataHandler( bool iWrite, VTableLookupRunParameter* iT = 0 );
	       ~VTableLookupDataHandler() {}
		
		bool cut()                                //!< apply cuts on successfull reconstruction to input data
		{
			return cut( false );
		}
		bool cut( bool bWrite );
		void fill();                              //!< fill output tree
		void fillMChistograms();
		void setfillTables( bool ib )
		{
			fwrite = ib;
		}
		int getArrayReconstructionMethod()
		{
			return fMethod;
		}
		double* getDistance()
		{
			return fdist;
		}
		double* getDistance( ULong64_t iTelType );
		double* getDistanceToCore()
		{
			return fR;
		}
		double* getDistanceToCore( ULong64_t iTelType, bool iMC = false );
		int    getEventNumber()
		{
			return eventNumber;
		}
		bool   getEventStatus()
		{
			return fEventStatus;
		}
		double getEventWeight()
		{
			return fEventWeight;
		}
		double getMCAz()
		{
			return fMCaz;
		}
		double getMCZe()
		{
			return fMCze;
		}
		double getAz()
		{
			return fAz;
		};
		double getMCDistance();
		double* getMCEnergyArray();
		double getMCEnergy()
		{
			return fMCEnergy;
		}
		double getMCWobbleOffset()
		{
			return sqrt( fMCxoff * fMCxoff + fMCyoff * fMCyoff );
		}
		int    getEventCounter()
		{
			return fEventCounter;
		}
                double* getFROGS_goodness();
                double* getFROGS_goodness( ULong64_t iTelType );
		double* getLength()
		{
			return flength;
		}
		double* getLength( ULong64_t iTelType );
		map<ULong64_t, unsigned int > getList_of_Tel_type()
		{
			return fList_of_Tel_type;
		}
                double* getLoss()
                {
                        return floss;
                }
                double* getLoss( ULong64_t iTelType );
		unsigned int getNTel_type( ULong64_t t )
		{
			if( fList_of_Tel_type.find( t ) != fList_of_Tel_type.end() )
			{
				return fList_of_Tel_type[t];
			}
			else
			{
				return 99999;
			}
		}
		unsigned int getMaxNbrTel() const
		{
			return VDST_MAXTELESCOPES;
		}
		int* getNtubes()
		{
			return fntubes;
		}
		double* getMSCWtel()
		{
			return ftmscw;
		}
		double* getMSCLtel()
		{
			return ftmscl;
		}
		double getMCMinEnergy()
		{
			return fMCMinEnergy;
		}
		double getMCMaxEnergy()
		{
			return fMCMaxEnergy;
		}
		double getMCSpectralIndex()
		{
			return fMCSpectralIndex;
		}
		double getSpectralIndex()
		{
			return fSpectralIndex;
		}
		int  getNEntries()
		{
			return fNEntries;
		}
		int  getNEvents()
		{
			if( fTshowerpars )
			{
				return fTshowerpars->GetEntries();
			}
			else
			{
				return 0;
			}
		}
		bool getNextEvent();                      //!< get next event from evndisp tree
		bool getNextEvent( bool bShort );         //!< get next event from evndisp tree
		double getMeanNoiseLevel( bool bCurrentNoiseLevel = false )
		{
			if( !bCurrentNoiseLevel )
			{
				return calculateMeanNoiseLevel( false );
			}
			else
			{
				return ( double )fmeanPedvar_Image;
			}
		}
		vector< double > getNoiseLevel( bool bCurrentNoiseLevel = false )
		{
			if( !bCurrentNoiseLevel )
			{
				return fNoiseLevel;
			}
			else
			{
				return fCurrentNoiseLevel;
			}
		}
		map< ULong64_t, double > getNoiseLevel_per_TelescopeType();
		unsigned int getNTel()
		{
			return fNTel;
		}
		unsigned int getNTelTypes()
		{
			return fNTelTypes;
		}
		TFile* getOutputFile()
		{
			return fOutFile;
		}
		double* getSize( double iSizeCorrection = 1., bool iSelectedImagesOnly = false, bool iSize2 = false );
		double* getSize( double iSizeCorrection, ULong64_t iTelType, bool iSelectedImagesOnly, bool iSize2 = false );
		double* getSize2( double iSizeCorrection = 1., bool iSelectedImagesOnly = false )
		{
			return getSize( iSizeCorrection, iSelectedImagesOnly, true );
		}
		double* getSize2( double iSizeCorrection, ULong64_t iTelType, bool iSelectedImagesOnly )
		{
			return getSize( iSizeCorrection, iTelType, iSelectedImagesOnly, true );
		}
                double* getTimeGradient()
                {
                        return ftgrad_x;
                }
                double* getTimeGradient( ULong64_t iTelType );
		double* getWeight()
		{
			return fweight;
		}
		double* getWidth()
		{
			return fwidth;
		}
		double* getWidth( ULong64_t iTelType );
		double getMaxTotalTime()
		{
			return fMaxTotalTime;
		}
		double getTelElevation( unsigned int iTelID )
		{
			return fTelElevation[iTelID];
		}
		double getTelElevation();
		ULong64_t getTelType( unsigned int iTelID )
		{
			return fTel_type[iTelID];
		}
		unsigned int getTelType_arraycounter( unsigned int iTelID );      // return position of tel type for this telescope in array counter
		double getZe();
		double getTheta2()
		{
			return ftheta2;
		}
		double getWobbleOffset()
		{
			return sqrt( fXoff * fXoff + fYoff * fYoff );
		}
		bool isReconstructed();
		bool readRunParameter();
		void reset();                             //!< reset a few output variables
		void resetAll();
		//!< no check of boundaries!!
		void setDistanceToCore( int itel, double iR )
		{
			fR[itel] = iR;
		}
		void setEnergy( double iES, double iChi2S, double idES )
		{
			fenergyS = iES;
                        fechi2S = iChi2S;
                        fdES = idES;
		}
		void setEnergyT( int i, double iETS )
		{
			fES[i] = iETS;
		}
		bool setInputFile( vector< string > );              //!< set input file
		void setMCMinEnergy( double iB )
		{
			fMCMinEnergy = iB;
		}
		void setMCMaxEnergy( double iB )
		{
			fMCMaxEnergy = iB;
		}
		void setMCSpectralIndex( double iB )
		{
			fMCSpectralIndex = iB;
		}
		void setSpectralIndex( double iB )
		{
			fSpectralIndex = iB;
		}
		void setMethod( int );                    //!< data from which evndisp reconstruction method should be used?
		void setMSCL( double iMSLC )
		{
			fmscl = iMSLC;
		}
		void setMSCLT( int i, double iMSLC, float iMSLC_T = -99. )
		{
			ftmscl[i] = iMSLC;
			ftmscl_sigma[i] = iMSLC_T;
		}
		void setNMSCW( int in )
		{
			fnmscw = in;
		}
		void setMSCW( double iMSWC )
		{
			fmscw = iMSWC;
		}
		void setMSCWT( int i, double iMSWC, float iMSWC_T = -99. )
		{
			ftmscw[i] = iMSWC;
			ftmscw_sigma[i] = iMSWC_T;
		}
		void setMSCT( double iMSWT )
		{
			fmsct = iMSWT;
		}
		void setMSCTT( int i, double iMSTC, float iMSTC_T = -99. )
		{
			ftmsct[i] = iMSTC;
			ftmsct_sigma[i] = iMSTC_T;
		}
		void setMSC_FRGO( double iMSW_FRGO )
		{
			fmsc_frgo = iMSW_FRGO;
		}
		void setMSC_FRGOT( int i, double iMSTC, float iMSTC_T = -99. )
		{
			ftmsc_frgo[i] = iMSTC;
			ftmsc_frgo_sigma[i] = iMSTC_T;
		}
		void setMWR( double iM )
		{
			fmwr = iM;
		}
		void setMLR( double iM )
		{
			fmlr = iM;
		}
		void setMCDistanceToCameraCenter( double iMin, double iMax )
		{
			fMC_distance_to_cameracenter_min = iMin;
			fMC_distance_to_cameracenter_max = iMax;
		}
		void setDebug( bool iD )
		{
			fDebug = iD;
		}
		bool setOutputFile( string, string, string );
		bool terminate( TNamed* );                //!< write everything to disk
};
#endif
