#!/bin/sh
#
# calculate effective areas and instrument response functions for CTA
#
#
# Author: Gernot Maier
#
##############################################################################


# if [ $# -lt 7 ]
#then
#   echo
#   echo "CTA.EFFAREA.sub_analyse_list.sh <subarray> <recid> <particle> <cutfile template>"
#   echo "                                <scripts input parameter file> <outputsubdirectory> <data set>"
#   echo "                                [filling mode] [qsub options] [direction (e.g. _180deg)]"
#   echo "================================================================================"
#   echo
#   echo "calculate effective areas and instrument response functions for CTA"
#   echo 
#   echo
#   echo "<subarray>"
#   echo "     subarray for analysis (e.g. array E)"
#   echo
#   echo "<recid>"
#   echo "     reconstruction ID from array reconstruction (mscw stage)"
#   echo
#   echo "<particle>"
#   echo "     allowed particle types are:"
#   echo "     ALL/onSource/cone"
#   echo
#   echo "<cutfile template>"
#   echo "     template for gamma/hadron cut file (full path and file name)"
#   echo "     (examples can be found in CTA_EVNDISP_AUX_DIR/GammaHadronCutFiles)"
#   echo
#   echo "<scripts parameter file>"
#   echo "     file with analysis parameter"
#   echo "     see e.g. CTA_EVNDISP_AUX_DIR/ParameterFiles/scriptsInput.runparameter"
#   echo 
#   echo "<outputsubdirectory>"
#   echo "     directory with all result (effective areas) and log files (full path)"
#   echo
#   echo "<data set>         e.g. cta-ultra3, ISDC3700m, ...  "
#   echo
#   echo "[filling mode]"
#   echo "     effective area filling mode (use 2 to calculate angular resolution only)"
#   echo "     (optional, for default full calculation no option is needed)"
#   echo
#   echo "[direction]        e.g. for north: \"_180deg\", for south: \"_0deg\", for all directions: no option"
#   echo
#   exit
#fi

echo
echo "calculating effective areas for CTA: create run scripts"
echo "------------------------------------------------"
echo

source $EVNDISPSYS/setObservatory.sh CTA

######################################################################
# input variables
######################################################################
ANAPAR=PPPANAPAR
ARRAY=PPPARRAY
RECID=PPPRECID
PTYPE=PPPVPART
CFIL=PPPCCUT
ODIR=PPPODIR
GFILLING=PPPGMOD
DSET=PPPDSET
MCAZ="PPPMCAZ"

#######################################
# read analysis values from parameter file
if [ ! -e $ANAPAR ]
then
  echo "error: analysis parameter file not found: $ANAPAR" 
  exit
fi
# TMPDIR="/lustre/fs9/group/cta/users/maierg/CTA//tmp/"
cp $ANAPAR $TMPDIR
ANAPARF=`basename $ANAPAR`
ANAPAR="${TMPDIR}/${ANAPARF}"
echo "reading analysis parameter from $ANAPAR"
NIMAGESMIN=`grep NIMAGESMIN $ANAPAR | awk {'print $2'}`
NTYPEMIN_0=`grep NTYPEMIN_0 $ANAPAR | awk {'print $2'}`
NTYPEMIN_1=`grep NTYPEMIN_1 $ANAPAR | awk {'print $2'}`
NTYPEMIN_2=`grep NTYPEMIN_2 $ANAPAR | awk {'print $2'}`
if [ -z "$NTYPEMIN_0" ]
then
   NTYPEMIN_0=0
fi
if [ -z "$NTYPEMIN_1" ]
then
   NTYPEMIN_1=0
fi
if [ -z "$NTYPEMIN_2" ]
then
   NTYPEMIN_2=0
fi

ANADIR=`grep MSCWSUBDIRECTORY  $ANAPAR | awk {'print $2'}`
TMVACUT=`grep TMVASUBDIR $ANAPAR | awk {'print $2'}`
EFFAREADIR=`grep EFFAREASUBDIR $ANAPAR | awk {'print $2'}`
OBSTIME=`grep OBSERVINGTIME_H $ANAPAR | awk {'print $2'}`
GETXOFFYOFFAFTERCUTS=`grep GETXOFFYOFFAFTERCUTS $ANAPAR | awk  {'print $2'}`
echo "Input parameters read from $ANAPAR"
echo $NIMAGESMIN $ANADIR $TMVACUT $EFFAREADIR $OBSTIME

if [ -z "$ANADIR" ] || [ -z "$NIMAGESMIN" ] || [ -z "$TMVACUT" ] || [ -z "$EFFAREADIR" ] || [ -z "$OBSTIME" ] || [ -z "$NTYPEMIN_0" ]
then
  echo "error: analysis parameter file not correct: $ANAPAR" 
  echo " one variable missing"
  exit
fi

######################################################################
# directories
######################################################################
echo "data (input) directory"
# if [ $PART = "gamma_onSource" ] || [ $PART = "electron_onSource" ] || [ $PART = "proton_onSource" ] 
#then
#  DDIR=$CTA_USER_DATA_DIR/analysis/AnalysisData/$DSET/$ARRAY/$ANADIR-onAxis/
#else
#  DDIR=$CTA_USER_DATA_DIR/analysis/AnalysisData/$DSET/$ARRAY/$ANADIR/
#fi
# CHANGE 2015-11-10: use always result files from off-axis lookup tables (prod3 only?)
DDIR=$CTA_USER_DATA_DIR/analysis/AnalysisData/$DSET/$ARRAY/$ANADIR/
echo $DDIR
mkdir -p $DDIR
echo "output data directory"
echo $ODIR
mkdir -p $ODIR

declare -a PTYPELIST=("gamma_onSource" "electron_onSource" "proton_onSource" "gamma_cone" "electron" "proton")
if [ $PTYPE = "GAMMA" ]
then
    declare -a PTYPELIST=("gamma_onSource" "gamma_cone")
fi

######################################################################
# loop over all particle types
######################################################################
for PART in "${PTYPELIST[@]}"
do
    echo "$PART"
    echo
    ######################################################################
    # input files and input parameters
    ######################################################################
    # parameters which are the same for all particle types
    # no AZ bins (CTA sims are sorted already on the input side in AZ)
    AZBINS="0"
    TELTYPECUTS="1"
    # data directory
    # on source gamma rays
    if [ $PART = "gamma_onSource" ]
    then
       MSCFILE=$DDIR/gamma_onSource."$ARRAY"_ID"$RECID$MCAZ"*.mscw.root
       EFFFILE=$DDIR/EffectiveAreas/
       OFIL=gamma_onSource."$ARRAY"_ID"$RECID".eff
       OFFMIN=( 0. )
       OFFMAX=( 100000. )
       # changed 2015-11-17: always use off-axis TMVA for gamma/hadron separation
       # OFFMEA=( "0.0" )
       OFFMEA=( "0.5" )
    # NOTE: this is theta2
       THETA2MIN=( -1. )
    # using TMVA or angular resolution
       THETA2MAX=( -1. )
       ISOTROPY="0"
       DIRECTIONCUT="2"
    fi
    # isotropic gamma-rays: analyse in rings in camera distance
    if [ $PART = "gamma_cone" ]
    then
       MSCFILE=$DDIR/gamma_cone."$ARRAY"_ID"$RECID$MCAZ"*.mscw.root
       EFFFILE=$DDIR/EffectiveAreas/
       OFIL=gamma_cone."$ARRAY"_ID"$RECID".eff
       # note that these bins are also hardwired in VTableLookupRunParameter::setCTA_MC_offaxisBins
       OFFMIN=( 0.0 1.0 2.0 3.0 4.0 5.0 )
       OFFMAX=( 1.0 2.0 3.0 4.0 5.0 6.0 )
       OFFMEA=( 0.5 1.5 2.5 3.5 4.5 5.5 )
    # NOTE: this is theta2
       THETA2MIN=( -1. )
       THETA2MAX=( -1. )
       ISOTROPY="0"
       DIRECTIONCUT="2"
    fi
    if [ $PART = "electron" ] || [ $PART = "electron_onSource" ]
    then
       MSCFILE=$DDIR/electron."$ARRAY"_ID"$RECID$MCAZ"*.mscw.root
       EFFFILE=$DDIR/EffectiveAreas/
       OFFMIN=( 0. )
       OFFMAX=( 100000. )
    # NOTE: this is theta and not theta2
       if [ $PART = "electron" ]
       then
          OFIL=electron."$ARRAY"_ID"$RECID".eff
          THETA2MIN=( 0.0 1.0 2.0 3.0 4.0 5.0 )
          THETA2MAX=( 1.0 2.0 3.0 4.0 5.0 6.0 )
          OFFMEA=( 0.5 1.5 2.5 3.5 4.5 5.5 )
       else
          OFIL=electron_onSource."$ARRAY"_ID"$RECID".eff
          THETA2MIN=( 0. )
          THETA2MAX=( 1. )
          # changed 2015-11-17: always use off-axis TMVA for gamma/hadron separation
          # OFFMEA=( "0.0" )
          OFFMEA=( 0.5 )
       fi   
       ISOTROPY="1"
       DIRECTIONCUT="0"
    fi
    if [ $PART = "proton" ] || [ $PART = "proton_onSource" ]
    then
       MSCFILE=$DDIR/proton*."$ARRAY"_ID"$RECID$MCAZ"*.mscw.root
       if [ $ARRAY = "V5" ]
       then
          MSCFILE=$DDIR/proton."$ARRAY"_ID"$RECID$MCAZ".mscw.root
       fi
       EFFFILE=$DDIR/EffectiveAreas/
       OFIL=proton."$ARRAY"_ID"$RECID".eff
       OFFMIN=( 0. )
       OFFMAX=( 100000. )
    # NOTE: this is theta and not theta2
       if [ $PART = "proton" ] 
       then
          OFIL=proton."$ARRAY"_ID"$RECID".eff
          THETA2MIN=( 0.0 1.0 2.0 3.0 4.0 5.0 )
          THETA2MAX=( 1.0 2.0 3.0 4.0 5.0 6.0 )
          OFFMEA=( 0.5 1.5 2.5 3.5 4.5 5.5 )
       else
          OFIL=proton_onSource."$ARRAY"_ID"$RECID".eff
          THETA2MIN=( 0. )
          THETA2MAX=( 2. )
    # VALID      THETA2MAX=( 1. )
          # changed 2015-11-17: always use off-axis TMVA for gamma/hadron separation
          # OFFMEA=( "0.0" )
          OFFMEA=( 0.5 )
       fi
       ISOTROPY="1"
       DIRECTIONCUT="0"
    fi
    NOFF=${#OFFMIN[@]}
    NTH2=${#THETA2MIN[@]}
    ######################################################################


    ###############################################################################
    # loop over all MC cuts
    for ((i=0; i < $NOFF; i++))
    do
       iMIN=${OFFMIN[$i]}
       iMAX=${OFFMAX[$i]}
    # loop over all theta2 cuts
       for ((j=0; j < $NTH2; j++))
       do
         jMIN=${THETA2MIN[$j]}
         jMAX=${THETA2MAX[$j]}

    ###############################################################################
    # theta2 cut of protons and electron should match the rings from the isotropic gammas
          if [ $PART = "proton" ] || [ $PART = "proton_onSource" ] || [ $PART = "electron" ] || [ $PART = "electron_onSource" ]
          then
             jMIN=$(echo "$jMIN*$jMIN" | bc -l )
             jMAX=$(echo "$jMAX*$jMAX" | bc -l )
          fi

    ###############################################################################
    # create cut file
          iCBFILE=`basename $CFIL`      
          if [ $PART = "gamma_onSource" ] || [ $PART = "gamma_cone" ] 
          then
              CFILP="${CFIL}.gamma.dat"
          else
              CFILP="${CFIL}.CRbck.dat"
          fi
          iCFIL=$ODIR/ANASUM.GammaHadron-$DSET-$PART-$i-$j-MCAZ${MCAZ}.$iCBFILE.dat

          if [ ! -e $CFILP ]
          then
            echo "ERROR: cut file does not exist:"
            echo $CFILP
            exit
          fi
          cp -f $CFILP $iCFIL

    # wobble offset
          if [ $PART = "gamma_onSource" ] || [ $PART = "gamma_cone" ] 
          then
             WOBBLEOFFSET=${OFFMEA[$i]}
          else
             WOBBLEOFFSET=${OFFMEA[$j]}
          fi
    # angular resolution file
          if [ $PART = "gamma_onSource" ] 
          then
             ANGRESFILE="$CTA_USER_DATA_DIR/analysis/AnalysisData/$DSET/$EFFAREADIR/AngularResolution/gamma_onSource."$ARRAY"_ID"$RECID".eff-0.root"
          else
             ANGRESFILE="$CTA_USER_DATA_DIR/analysis/AnalysisData/$DSET/$EFFAREADIR/AngularResolution/gamma_cone."$ARRAY"_ID"$RECID".eff-$i.root"
          fi
    # particle number file
          if [ $PART = "gamma_onSource" ] || [ $PART = "electron_onSource" ] || [ $PART = "proton_onSource" ]
          then
             PNF="$CTA_USER_DATA_DIR/analysis/AnalysisData/$DSET/$EFFAREADIR/QualityCuts001CU/ParticleNumbers."$ARRAY".00.root"
          elif [ $PART = "gamma_cone" ]
          then
             PNF="$CTA_USER_DATA_DIR/analysis/AnalysisData/$DSET/$EFFAREADIR/QualityCuts001CU/ParticleNumbers."$ARRAY".$i.root"
          else
             PNF="$CTA_USER_DATA_DIR/analysis/AnalysisData/$DSET/$EFFAREADIR/QualityCuts001CU/ParticleNumbers."$ARRAY".$j.root"
          fi
          

          sed -i -e "s|OFFMIN|$iMIN|" \
                 -e "s|OFFMAX|$iMAX|" \
                 -e "s|THETA2MIN|$jMIN|" \
                 -e "s|THETA2MAX|$jMAX|" \
                 -e "s|DIRECTIONCUT|$DIRECTIONCUT|" \
                 -e "s|SUBARRAY|$ARRAY|" \
                 -e "s|MINIMAGES|$NIMAGESMIN|" \
                 -e "s|NTELTYPE0|$NTYPEMIN_0|" \
                 -e "s|NTELTYPE1|$NTYPEMIN_1|" \
                 -e "s|NTELTYPE2|$NTYPEMIN_2|" \
                 -e "s|WOBBLEOFFSET|$WOBBLEOFFSET|" \
                 -e "s|TMVACUTDIR|$TMVACUT|" \
                 -e "s|DATASET|$DSET|" \
                 -e "s|ANGRESFILE|$ANGRESFILE|" \
                 -e "s|PARTICLENUMBERFILE|$PNF|" \
                 -e "s|OBSERVINGTIME_H|$OBSTIME|" $iCFIL

          cd $EVNDISPSYS/scripts/CTA/
          echo $iCFIL

    ###############################################################################
    # create run list
          MSCF=$ODIR/effectiveArea-CTA-$DSET-$PART-$i-$j.$ARRAY-MCAZ${MCAZ}.dat
          rm -f $MSCF
          echo "effective area data file for $PART $i $j" > $MSCF
    ###############################################################################
    # general run parameters
    ###############################################################################
    # filling mode
    ###############################################################################
    # fill IRFs and effective areas
          if [ $PART = "gamma_onSource" ] || [ $PART = "gamma_cone" ]
          then
    # filling mode 0: fill and use angular resolution for energy dependent theta2 cuts
             echo "* FILLINGMODE $GFILLING" >> $MSCF
          else
    # background: use fixed theta2 cut
             echo "* FILLINGMODE 3" >> $MSCF
          fi
    # fill IRFs only
          echo "* ENERGYRECONSTRUCTIONMETHOD 1" >> $MSCF
          echo "* ENERGYAXISBINS 60" >> $MSCF
          echo "* ENERGYRECONSTRUCTIONQUALITY 0" >> $MSCF
    # one azimuth bin only
          echo "* AZIMUTHBINS $AZBINS" >> $MSCF
          echo "* ISOTROPICARRIVALDIRECTIONS $ISOTROPY" >> $MSCF
          echo "* TELESCOPETYPECUTS $TELTYPECUTS" >> $MSCF
    # do fill analysis (a 1 would mean that MC histograms would be filled only)
          echo "* FILLMONTECARLOHISTOS 0" >> $MSCF
    # spectral index & CR spectra
          if [ $PART = "proton" ] || [ $PART = "proton_onSource" ]
          then
             echo "* ENERGYSPECTRUMINDEX  1 2.5 0.1" >> $MSCF
             echo "* ESPECTRUM_FOR_WEIGHTING $CTA_EVNDISP_AUX_DIR/AstroData/TeV_data/EnergySpectrum_literatureValues_CR.dat 0" >> $MSCF
             if  [ $GETXOFFYOFFAFTERCUTS = "yes" ]
             then	
                 echo "* GETXOFFYOFFAFTERCUTS 1" >> $MSCF
             fi    

          fi
          if [ $PART = "electron" ] || [ $PART = "electron_onSource" ]
          then
             echo "* ENERGYSPECTRUMINDEX  1 2.5 0.1" >> $MSCF
             echo "* ESPECTRUM_FOR_WEIGHTING $CTA_EVNDISP_AUX_DIR/AstroData/TeV_data/EnergySpectrum_literatureValues_CR.dat 8" >> $MSCF
             if  [ $GETXOFFYOFFAFTERCUTS = "yes" ]
             then
             echo "* GETXOFFYOFFAFTERCUTS 1" >> $MSCF
             fi
          fi
          if [ $PART = "gamma_onSource" ] || [ $PART = "gamma_cone" ]
          then
             echo "* ENERGYSPECTRUMINDEX  1 2.5 0.1" >> $MSCF
             echo "* ESPECTRUM_FOR_WEIGHTING $CTA_EVNDISP_AUX_DIR/AstroData/TeV_data/EnergySpectrum_literatureValues_CrabNebula.dat 5" >> $MSCF
          fi
    # first half of data set is not used (as these events are used for the TMVA training)
          if [ $PART = "gamma_onSource" ] || [ $PART = "gamma_cone" ]
          then
             echo "* IGNOREFRACTIONOFEVENTS 0.5" >> $MSCF
          fi
          if [ $PART = "proton" ] || [ $PART = "proton_onSource" ]
          then
    # first half of data set is not used (as these events are used for the TMVA training)
             echo "* IGNOREFRACTIONOFEVENTS 0.5" >> $MSCF
          fi
          echo "* CUTFILE $iCFIL" >> $MSCF
          echo "* SIMULATIONFILE_DATA $MSCFILE" >> $MSCF

    # output file
          if [ $PART = "gamma_onSource" ] || [ $PART = "gamma_cone" ]
          then
             OFIX=$ODIR/$OFIL-$i
          else
             OFIX=$ODIR/$OFIL-$j
          fi

          echo
          echo "preparing new analysis run"
          echo "--------------------------"
          echo
          echo "gamma/hadron separation file"
          echo $iCFIL
          echo $PNF

      ##############################
      # run effective area code
          $EVNDISPSYS/bin/makeEffectiveArea $MSCF $OFIX.root > $OFIX.log

       done
    done
done

exit
