#!/bin/bash
# script to analyse data files with anasum (parallel analysis) from a simple run list

# EventDisplay version
EDVERSION=`$EVNDISPSYS/bin/anasum --version | tr -d .`

if [[ "$#" -lt 4 ]]; then
# begin help message
echo "
ANASUM parallel data analysis: submit jobs using a simple run list

ANALYSIS.anasum_parallel_from_runlist.sh <run list> <output directory> <cut set> <background model> [run parameter file] [mscw directory] [sim type] [method] [force atmosphere] [RECID]

required parameters:

    <run list>              simple runlist with a single run number per line
        
    <output directory>      anasum output files are written to this directory
                        
    <cut set>               hardcoded cut sets predefined in the script
                            (e.g., soft2tel, moderate3tel, etc.)
    
    <background model>      background model
                            (RE = reflected region, RB = ring background)
    
optional parameters:

    [run parameter file]    anasum run parameter file (located in 
                            \$VERITAS_EVNDISP_AUX_DIR/ParameterFiles/;
                            default is ANASUM.runparameter)

    [mscw directory]        directory containing the mscw.root files.
			    Default: $VERITAS_USER_DATA_DIR/analysis/Results/$EDVERSION

    [sim type]              use IRFs derived from this simulation type (GRISU-SW6 or CARE_June1425)
			    Default: CARE_June1425

    [method]                reconstruction method: GEO, DISP, FROGS.
			    Default: GEO

    [force atmosphere]	    use EAs generated with this atmospheric model (21 or 22).
			    Default: Atmosphere determined from run date for each run.				
			    Attention: Must use the same atmospere for EAs as was used for the lookup tables in the mscw_energy stage!

IMPORTANT! Run ANALYSIS.anasum_combine.sh once all parallel jobs have finished!

--------------------------------------------------------------------------------
"
#end help message
exit
fi

###########################
# IRFs
IRFVERSION="v500-pre05"
AUXVERSION="auxv01"

# Run init script
bash $(dirname "$0")"/helper_scripts/UTILITY.script_init.sh"
[[ $? != "0" ]] && exit 1

# Parse command line arguments
RLIST=$1
ODIR=$2
CUTS=$3
BACKGND=$4
[[ "$5" ]] && RUNP=$5  || RUNP="ANASUM.runparameter"
[[ "$6" ]] && INDIR=$6 || INDIR="$VERITAS_USER_DATA_DIR/analysis/Results/$EDVERSION/"
[[ "$7" ]] && SIMTYPE=$7 || SIMTYPE="CARE_Sep0416"
[[ "$8" ]] && METH=$8 || METH="GEO"
[[ "$9" ]] && FORCEDATMO=$9 
[[ "${10}" ]] && RECID=${10}

# cut definitions (note: VX to be replaced later in script)
if [[ $CUTS == *superhard* ]]; then
    CUT="NTel3-PointSource-SuperHard"
elif [[ $CUTS == *supersoft* ]]; then
    CUT="NTel2-PointSource-SuperSoftSpectrum"
elif [[ $CUTS == *moderateopen* ]]; then
    CUT="NTel2-PointSource-ModerateOpen"
elif [[ $CUTS == *softopen* ]]; then
    CUT="NTel2-PointSource-SoftOpen"
elif [[ $CUTS == *hardopen* ]]; then
    CUT="NTel2-PointSource-HardOpen"
elif [[ $CUTS == *softExt* ]]; then
    CUT="NTel2-ExtendedSource-Soft"
elif [[ $CUTS == *soft2tel* ]]; then
    CUT="NTel2-PointSource-Soft"
elif [[ $CUTS = *moderate2tel* ]]; then
    CUT="NTel2-PointSource-Moderate"
elif [[ $CUTS = *moderateExt2tel* ]]; then
    CUT="NTel2-ExtendedSource-Moderate"
elif [[ $CUTS = *moderateExt3tel* ]]; then
    CUT="NTel3-ExtendedSource-Moderate"
elif [[ $CUTS = *moderate3tel* ]]; then
    CUT="NTel3-PointSource-Moderate"
elif [[ $CUTS = *hard2tel* ]]; then
    CUT="NTel2-PointSource-Hard"
elif [[ $CUTS = *hard3tel* ]]; then
    CUT="NTel3-PointSource-Hard"
elif [[ $CUTS = *hard4tel* ]]; then
    CUT="NTel4-PointSource-Hard"
elif [[ $CUTS = *hardExt2tel* ]]; then
    CUT="NTel2-ExtendedSource-Hard"
elif [[ $CUTS = *hardExt3tel* ]]; then
    CUT="NTel3-ExtendedSource-Hard"
elif [[ $CUTS = *frogs25* ]]; then
    CUT="FROGS_NTel2_001-003-005CU_index2.5"
elif [[ $CUTS = *frogs43* ]]; then
    CUT="FROGS_NTel2_001-003-005CU_index4.3"
elif [[ $CUTS = *BDTmoderate* ]]; then
    CUT="NTel2-PointSource-Moderate-TMVA-BDT"
elif [[ $CUTS = *BDTsoft* ]]; then
    CUT="NTel2-PointSource-Soft-TMVA-BDT"
elif [[ $CUTS = *BDThard* ]]; then
    CUT="NTel2-PointSource-Hard-TMVA-BDT"
else
    echo "ERROR: unknown cut definition: $CUTS"
    exit 1
fi
CUTFILE="ANASUM.GammaHadron-Cut-${CUT}.dat"
EFFAREA="effArea-${IRFVERSION}-${AUXVERSION}-${SIMTYPE}-Cut-${CUT}-${METH}-VX-ATMXX-TX.root"
# START TEMPORARY
EFFAREA="effArea-${IRFVERSION}-${AUXVERSION}-${SIMTYPE}-Cut-${CUT}-${METH}ID${RECID}-VX-ATMXX-TX.root"
# END TEMPORARY
if [[ "$CUT" == *PointSource* ]]; then
    CUTRED="${CUT//-PointSource-/-}"
fi
if [[ "$CUT" == *ExtendedSource* ]]; then
    CUTRED="${CUT//-ExtendedSource-/-}"
fi
RADACC="radialAcceptance-${IRFVERSION}-${AUXVERSION}-${SIMTYPE}-Cut-${CUTRED}-${METH}-VX-TX.root"
# START TEMPORARY
RADACC="IGNOREACCEPTANCE"
# END TEMPORARY

echo $CUTFILE
echo $EFFAREA
echo $RADACC

# background model parameters
if [[ "$BACKGND" == *RB* ]]; then
    BM="RB"
    BMPARAMS="0.6 20"
elif [[ "$BACKGND" == *RE* ]]; then
    BM="RE"
    BMPARAMS="0.1 2 6"
else
    echo "ERROR: unknown background model: $BACKGND"
    echo "Allowed values are: RE, RB"
    exit 1
fi

# Check that run list exists
if [[ ! -f "$RLIST" ]]; then
    echo "Error, simple runlist $RLIST not found, exiting..."
    exit 1
fi

# Check that run parameter file exists
if [[ "$RUNP" == `basename $RUNP` ]]; then
    RUNP="$VERITAS_EVNDISP_AUX_DIR/ParameterFiles/$RUNP"
fi
if [ ! -f "$RUNP" ]; then
    echo "Error, anasum run parameter file not found, exiting..."
    echo "(searched for $RUNP)"
    exit 1
fi

# directory for run scripts
DATE=`date +"%y%m%d"`
LOGDIR="$VERITAS_USER_LOG_DIR/$DATE/ANASUM.ANADATA"
echo -e "Log files will be written to:\n $LOGDIR"
mkdir -p $LOGDIR

# output directory for anasum products
echo -e "Output files will be written to:\n $ODIR"
mkdir -p $ODIR

#########################################
# make anasum run list
ANARUNLIST="$ODIR/$CUTS.anasum.dat"
rm -f $ANARUNLIST
echo "anasum run list: $ANARUNLIST"

# run list header
echo "* VERSION 6" >> $ANARUNLIST
echo "" >> $ANARUNLIST

RUNS=`cat $RLIST`

for RUN in ${RUNS[@]}; do
    # get array epoch, atmosphere and telescope combination for this run
    RUNINFO=`$EVNDISPSYS/bin/printRunParameter $INDIR/$RUN.mscw.root -runinfo`
    EPOCH=`echo $RUNINFO | awk '{print $(1)}'`
    ATMO=${FORCEDATMO:-`echo $RUNINFO | awk '{print $(2)}'`}
    if [[ $ATMO == *error* ]]; then
       echo "error finding atmosphere; skipping run $RUN"
       continue
    fi
    TELTOANA=`echo $RUNINFO | awk '{print "T"$(4)}'`
    echo "RUN $RUN at epoch $EPOCH and atmosphere $ATMO (Telescopes $TELTOANA)"
    echo "File $INDIR/$RUN.mscw.root"
    
    # do string replacements
    EFFAREARUN=${EFFAREA/VX/$EPOCH}
    EFFAREARUN=${EFFAREARUN/TX/$TELTOANA}
    EFFAREARUN=${EFFAREARUN/XX/$ATMO}
    RADACCRUN=${RADACC/VX/$EPOCH}
    RADACCRUN=${RADACCRUN/TX/$TELTOANA}

    if [[ $CUTS = *frogs* ]]; then
       RADACCRUN="radialAcceptance-v470-auxv01-Cut-NTel2-Moderate-GEO-V6-T1234.root"
    fi
    RADACC="IGNOREACCEPTANCE"

    
    # write line to file (VERSION 6)
    echo "* $RUN $RUN 0 $CUTFILE $BM $EFFAREARUN $BMPARAMS $RADACCRUN" >> $ANARUNLIST
    echo "* $RUN $RUN 0 $CUTFILE $BM $EFFAREARUN $BMPARAMS $RADACCRUN"
    # write line to file
#    echo "* $RUN $RUN 0 $BM $EFFAREARUN $BMPARAMS $RADACCRUN" >> $ANARUNLIST
#    echo "* $RUN $RUN 0 $BM $EFFAREARUN $BMPARAMS $RADACCRUN"
done

# submit the job
SUBSCRIPT="$EVNDISPSYS/scripts/VTS/ANALYSIS.anasum_parallel"
$SUBSCRIPT.sh $ANARUNLIST $INDIR $ODIR $RUNP

exit
