#!/bin/bash
# IRF production script (VERITAS)
#

if [ $# -lt 2 ]; then
# begin help message
echo "
IRF generation: produce a full set of instrument response functions (IRFs)

IRF.production.sh <sim type> <IRF type> [epoch] [atmosphere] [Rec ID] [BDT cuts] [cuts list file] [sim directory]

required parameters:

    <sim type>              original VBF file simulation type (e.g. GRISU-SW6, CARE_June1425)
    
    <IRF type>              type of instrument response function to produce
                            (e.g. EVNDISP, MAKETABLES, COMBINETABLES,
                             ANALYSETABLES, EFFECTIVEAREAS, COMBINEEFFECTIVEAREAS,
                             MVAEVNDISP, TRAINMVANGRES )
    
optional parameters:
    
    [epoch]                 array epoch(s) (e.g., V4, V5, V6)
                            (default: \"V4 V5 V6\")
                            
    [atmosphere]            atmosphere model(s) (21 = winter, 22 = summer)
                            (default: \"21 22\")
                            
    [Rec ID]                reconstruction ID(s) (default: \"0 2 3 4 5\")
                            (see EVNDISP.reconstruction.runparameter)

    [BDT cuts]              using cuts list for BDT cuts (e.g. 0=not used, 1=used)
                            (default: \"0\")
    
    [cuts list file]        file containing one gamma/hadron cuts file per line
                            (default: hard-coded standard EventDisplay cuts)

    [sim directory]         directory containing simulation VBF files

--------------------------------------------------------------------------------
"
#end help message
exit
fi

# Run init script
bash $(dirname "$0")"/helper_scripts/UTILITY.script_init.sh"
[[ $? != "0" ]] && exit 1

# Parse command line arguments
SIMTYPE=$1
IRFTYPE=$2
[[ "$3" ]] && EPOCH=$3 || EPOCH="V6 V5 V4"
[[ "$4" ]] && ATMOS=$4 || ATMOS="21 22"
# [[ "$5" ]] && RECID=$5 || RECID="0 2 3 4 5"
[[ "$5" ]] && RECID=$5 || RECID="0"
[[ "$6" ]] && BDTCUTS=$6 || BDTCUTS="0"
[[ "$7" ]] && CUTSLISTFILE=$7 || CUTSLISTFILE=""
[[ "$8" ]] && SIMDIR=$8 || SIMDIR=""

# evndisplay version
EDVERSION=`$EVNDISPSYS/bin/printRunParameter --version | tr -d .`
# version string for aux files
AUX="auxv01"

# simulation types and definition of parameter space
if [[ ${SIMTYPE:0:5} = "GRISU" ]]; then
    # GrISU simulation parameters
    ZENITH_ANGLES=( 00 20 30 35 40 45 50 55 60 65 )
    NSB_LEVELS=( 075 100 150 200 250 325 425 550 750 1000 )
    WOBBLE_OFFSETS=( 0.5 0.00 0.25 0.75 1.00 1.25 1.50 1.75 2.00 )
    ZENITH_ANGLES=( 20 30 )
    NSB_LEVELS=( 150 250 )
    WOBBLE_OFFSETS=( 0.5 )
    if [[ $IRFTYPE == "MVAEVNDISP" ]]; then
       NSB_LEVELS=( 200 )
       WOBBLE_OFFSETS=( 0.5 )
    fi
    if [[ $IRFTYPE == "MVAEVNDISP" ]]; then
       NSB_LEVELS=( 200 )
       WOBBLE_OFFSETS=( 0.5 )
    fi
elif [ ${SIMTYPE:0:4} = "CARE" ]; then
    # CARE simulation parameters
    ZENITH_ANGLES=( 00 20 30 35 40 45 50 55 60 65 )
    NSB_LEVELS=( 50 80 120 170 230 290 370 450 )
    WOBBLE_OFFSETS=( 0.5 )
    NSB_LEVELS=( 160 200 250 )
    ZENITH_ANGLES=( 00 20 30 )
    ZENITH_ANGLES=( 00 20 30 35 40 45 50 55 60 65 )
    NSB_LEVELS=( 50 75 100 130 160 200 250 300 350 400 450 )
    if [[ $IRFTYPE == "MVAEVNDISP" ]]; then
       NSB_LEVELS=( 170 )
       WOBBLE_OFFSETS=( 0.5 )
    fi
else
    echo "Invalid simulation type. Exiting..."
    exit 1
fi

# Set gamma/hadron cuts
if [[ $CUTSLISTFILE != "" ]]; then
    if [ ! -f $CUTSLISTFILE ]; then
        echo "Error, cuts list file not found, exiting..."
        exit 1
    fi
    # read file containing list of cuts
    CUTLIST=$(IFS=$'\r\n'; cat $CUTSLISTFILE)
else
    if [[ $BDTCUTS == "0"  ]]; then
    # default list of cuts
        CUTLIST="ANASUM.GammaHadron-Cut-NTel2-PointSource-ModerateOpen.dat"
#        CUTLIST="ANASUM.GammaHadron-Cut-NTel3-PointSource-Moderate.dat 
#                 ANASUM.GammaHadron-Cut-NTel2-PointSource-Moderate.dat
#                 ANASUM.GammaHadron-Cut-NTel3-PointSource-SuperHard.dat
#                 ANASUM.GammaHadron-Cut-NTel3-PointSource-Hard.dat
#                 ANASUM.GammaHadron-Cut-NTel2-PointSource-Soft.dat"
#                 ANASUM.GammaHadron-Cut-NTel3-PointSource-Moderate200.dat
#                 ANASUM.GammaHadron-Cut-NTel3-PointSource-Moderate250.dat
#                 ANASUM.GammaHadron-Cut-NTel3-PointSource-Moderate300.dat 
#                 ANASUM.GammaHadron-Cut-NTel2-PointSource-Moderate.dat 
#                 ANASUM.GammaHadron-Cut-NTel3-PointSource-SuperHard.dat
#                 ANASUM.GammaHadron-Cut-NTel3-PointSource-Hard.dat
#                 ANASUM.GammaHadron-Cut-NTel2-PointSource-Moderate250.dat
#                 ANASUM.GammaHadron-Cut-NTel2-PointSource-Moderate200.dat
#                 ANASUM.GammaHadron-Cut-NTel2-PointSource-Moderate300.dat
#                 ANASUM.GammaHadron-Cut-NTel3-PointSource-Hard250.dat
#                 ANASUM.GammaHadron-Cut-NTel3-PointSource-Hard200.dat
#                 ANASUM.GammaHadron-Cut-NTel3-PointSource-Hard300.dat
#                 ANASUM.GammaHadron-Cut-NTel4-PointSource-Hard.dat
#                 ANASUM.GammaHadron-Cut-NTel4-PointSource-Hard250.dat
#                 ANASUM.GammaHadron-Cut-NTel4-PointSource-Hard200.dat
#                 ANASUM.GammaHadron-Cut-NTel4-PointSource-Hard300.dat"
#                 ANASUM.GammaHadron-Cut-NTel2-PointSource-ModerateOpen.dat
#                 ANASUM.GammaHadron-Cut-NTel2-PointSource-SoftOpen.dat"
#                 ANASUM.GammaHadron-Cut-NTel2-PointSource-HardOpen.dat 
#                 ANASUM.GammaHadron-Cut-NTel2-ExtendedSource-Moderate.dat 
#                 ANASUM.GammaHadron-Cut-NTel2-ExtendedSource-Soft.dat 
#                 ANASUM.GammaHadron-Cut-NTel2-ExtendedSource-Hard.dat 
#                 ANASUM.GammaHadron-Cut-NTel3-PointSource-Moderate.dat 
#                 ANASUM.GammaHadron-Cut-NTel3-PointSource-Hard.dat
#                 ANASUM.GammaHadron-Cut-NTel3-ExtendedSource-SuperHard.dat" 
#       CUTLIST="ANASUM.GammaHadron-Cut-NTel2-PointSource-Moderate.dat"
    else
    #BDT TMVA list of cuts
        CUTLIST="ANASUM.GammaHadron-Cut-NTel2-PointSource-Moderate-TMVA-BDT.dat 
                 ANASUM.GammaHadron-Cut-NTel2-PointSource-Soft-TMVA-BDT.dat 
                 ANASUM.GammaHadron-Cut-NTel2-PointSource-Hard-TMVA-BDT.dat" 
#       CUTLIST="ANASUM.GammaHadron-Cut-NTel2-PointSource-Moderate.dat"
    fi
fi
CUTLIST=`echo $CUTLIST |tr '\r' ' '`

############################################################
# loop over complete parameter space and submit production
for VX in $EPOCH; do
    for ATM in $ATMOS; do
       ######################
       # set lookup table file name
       TABLECOM="table-${EDVERSION}-${AUX}-${SIMTYPE}-ATM${ATM}-${VX}-"
       ######################
       # combine lookup tables
       if [[ $IRFTYPE == "COMBINETABLES" ]]; then
            TFIL="${TABLECOM}"
            for ID in $RECID; do
                echo "combine lookup tables"
                METH="RECMETHOD${ID}"
                ./IRF.combine_lookup_table_parts.sh "${TFIL}${METH}" $VX $ATM $ID $SIMTYPE 
            done
            continue
       fi
       ######################
       # combine effective areas
       if [[ $IRFTYPE == "COMBINEEFFECTIVEAREAS" ]]; then
            for ID in $RECID; do
                for CUTS in ${CUTLIST[@]}; do
                    echo "combine effective areas $CUTS"
                   ./IRF.combine_effective_area_parts.sh $CUTS $VX $ATM $ID $SIMTYPE $AUX
                done # cuts
            done
            continue
        fi
        for ZA in ${ZENITH_ANGLES[@]}; do
            ######################
            # train MVA for angular resolution
            if [[ $IRFTYPE == "TRAINMVANGRES" ]]; then
               if [[ ${SIMTYPE:0:5} = "GRISU" ]]; then
                   ./IRF.trainTMVAforAngularReconstruction.sh $VX $ATM $ZA 200 $SIMTYPE
               elif [[ ${SIMTYPE:0:4} = "CARE" ]]; then
                   ./IRF.trainTMVAforAngularReconstruction.sh $VX $ATM $ZA 170 $SIMTYPE
               fi
               continue
            fi
            for NOISE in ${NSB_LEVELS[@]}; do
                for WOBBLE in ${WOBBLE_OFFSETS[@]}; do
                    echo "Now processing epoch $VX, atmo $ATM, zenith angle $ZA, wobble $WOBBLE, noise level $NOISE"
                    ######################
                    # run simulations through evndisp
                    if [[ $IRFTYPE == "EVNDISP" ]] || [[ $IRFTYPE == "MVAEVNDISP" ]]; then
                       if [[ ${SIMTYPE:0:5} = "GRISU" ]]; then
                          SIMDIR=$VERITAS_DATA_DIR/simulations/"$VX"_FLWO/grisu/ATM"$ATM"
                       elif [[ ${SIMTYPE:0:12} = "CARE_Oct1402" ]]; then
                          SIMDIR=$VERITAS_DATA_DIR/simulations/"$VX"_FLWO/CARE_Oct1402/
                       elif [[ ${SIMTYPE:0:12} = "CARE_May1515" ]]; then
                          SIMDIR=$VERITAS_DATA_DIR/simulations/"$VX"_FLWO/CARE_May1515/
                       elif [[ ${SIMTYPE:0:12} = "CARE_Jun0415" ]]; then
                          SIMDIR=$VERITAS_DATA_DIR/simulations/"$VX"_FLWO/CARE_Jun0415/
                       elif [[ ${SIMTYPE:0:12} = "CARE_Jun1215" ]]; then
                          SIMDIR=$VERITAS_DATA_DIR/simulations/"$VX"_FLWO/CARE_Jun1215/
                       elif [[ ${SIMTYPE:0:12} = "CARE_Jun2015" ]]; then
                          SIMDIR=$VERITAS_DATA_DIR/simulations/"$VX"_FLWO/CARE_Jun2015/
                       elif [[ ${SIMTYPE:0:12} = "CARE_Sep0416" ]]; then
                          SIMDIR=$VERITAS_DATA_DIR/simulations/"$VX"_FLWO/CARE_Sep0416/
                       elif [[ ${SIMTYPE:0:4} = "CARE" ]]; then
                          SIMDIR=$VERITAS_DATA_DIR/simulations/"$VX"_FLWO/${SIMTYPE}
                       fi
                        ./IRF.evndisp_MC.sh $SIMDIR $VX $ATM $ZA $WOBBLE $NOISE $SIMTYPE
                    ######################
                    # make tables
                    elif [[ $IRFTYPE == "MAKETABLES" ]]; then
                        for ID in $RECID; do
                           ./IRF.generate_lookup_table_parts.sh $VX $ATM $ZA $WOBBLE $NOISE $ID $SIMTYPE
                        done #recid
                    ######################
                    # analyse table files
                    elif [[ $IRFTYPE == "ANALYSETABLES" ]]; then
                        TFIL="${TABLECOM}"
                        for ID in $RECID; do
                            METH="RECMETHOD${ID}"
                            ./IRF.mscw_energy_MC.sh "${TFIL}${METH}" $VX $ATM $ZA $WOBBLE $NOISE $ID $SIMTYPE
                        done
                    ######################
                    # analyse effective areas
                    elif [[ $IRFTYPE == "EFFECTIVEAREAS" ]]; then
                        for ID in $RECID; do
                            ./IRF.generate_effective_area_parts.sh "$CUTLIST" $VX $ATM $ZA $WOBBLE $NOISE $ID $SIMTYPE
                        done #recID
                    fi
                done #wobble
            done #noise
        done #ZA
    done #ATM
done  #VX

exit

