/*! \class VEffectiveAreaCalculator
 *  \brief calculate effective areas and energy spectra
 *
 */

#include "VEffectiveAreaCalculator.h"

/*!
 *  CALLED FOR CALCULATION OF EFFECTIVE AREAS
 *
 *  this constructor is called for FILLING of the effective area tree
 *  (calculation of effective areas)
 *
 */
VEffectiveAreaCalculator::VEffectiveAreaCalculator( VInstrumentResponseFunctionRunParameter* iRunPara, VGammaHadronCuts* icuts )
{
    fRunPara = iRunPara;
    if( !fRunPara )
    {
        cout << "VEffectiveAreaCalculator: no run parameters given" << endl;
        cout << "...exiting..." << endl;
        exit( EXIT_FAILURE );
    }
    reset();
    
    // no effective area file present
    bNOFILE = true;
    
    // number of bins for histograms
    nbins = fRunPara->fEnergyAxisBins_log10;
    
    // cuts
    fCuts = icuts;
    fZe.push_back( fRunPara->fze );
    fAreaRadius.push_back( fRunPara->fCoreScatterRadius );
    fScatterMode.push_back( fRunPara->fCoreScatterMode );
    for( unsigned int i = 0; i < fZe.size(); i++ )
    {
        fXWobble.push_back( 0. );
        fYWobble.push_back( 0. );
        fNoise.push_back( 0 );
        fPedVar.push_back( 0. );
    }
    setIgnoreEnergyReconstructionCuts( fRunPara->fIgnoreEnergyReconstructionQuality );
    setIsotropicArrivalDirections( fRunPara->fIsotropicArrivalDirections );
    setTelescopeTypeCuts( fRunPara->fTelescopeTypeCuts );
    setWobbleOffset( fRunPara->fXoff, fRunPara->fYoff );
    setNoiseLevel( fRunPara->fnoise, fRunPara->fpedvar );
    
    // spectral weighting class
    fSpectralWeight = new VSpectralWeight();
    setMonteCarloEnergyRange( fRunPara->fMCEnergy_min, fRunPara->fMCEnergy_max, TMath::Abs( fRunPara->fMCEnergy_index ) );
    
    // define output tree (all histograms are written to this tree)
    hisTreeList = new TList();
    // same list, but histograms only
    hisTreeListofHistograms = new TList();
    // list for temporary histograms
    hisVList    = new TList();
    
    // these histograms are filled into the output tree
    char hname[400];
    char htitle[400];
    
    // define range and binning of energy axis of effective area plots
    cout << "histogram parameters (bins, log10(Emin), log10(Emax)): " << nbins;
    cout << ", " << fEnergyAxis_minimum_defaultValue << ", " << fEnergyAxis_maximum_defaultValue << endl;
    sprintf( htitle, "title" );
    
    sprintf( hname, "hEmc" );
    hEmc = new TH1D( hname, "energy spectrum", nbins, fEnergyAxis_minimum_defaultValue, fEnergyAxis_maximum_defaultValue );
    hEmc->Sumw2();
    hEmc->SetXTitle( "energy_{MC} [TeV]" );
    hEmc->SetYTitle( "entries" );
    hisTreeList->Add( hEmc );
    hisTreeListofHistograms->Add( hEmc );
    
    sprintf( hname, "hEcut" );
    hEcut = new TH1D( hname, "energy spectrum, after cuts", nbins, fEnergyAxis_minimum_defaultValue, fEnergyAxis_maximum_defaultValue );
    hEcut->Sumw2();
    hEcut->SetXTitle( "energy_{MC} [TeV]" );
    hEcut->SetYTitle( "entries" );
    hisTreeList->Add( hEcut );
    hisTreeListofHistograms->Add( hEcut );
    
    sprintf( hname, "hEcutUW" );
    hEcutUW = new TH1D( hname, "unweighted energy spectrum, after cuts", nbins, fEnergyAxis_minimum_defaultValue, fEnergyAxis_maximum_defaultValue );
    hEcutUW->Sumw2();
    hEcutUW->SetXTitle( "energy_{MC} [TeV]" );
    hEcutUW->SetYTitle( "entries (unweighted)" );
    hisTreeList->Add( hEcutUW );
    hisTreeListofHistograms->Add( hEcutUW );
    
    sprintf( hname, "hEcutNoTh2" );
    hEcutNoTh2 = new TH1D( hname, "energy spectrum, no direction cut, after cuts", nbins, fEnergyAxis_minimum_defaultValue, fEnergyAxis_maximum_defaultValue );
    hEcutNoTh2->Sumw2();
    hEcutNoTh2->SetXTitle( "energy_{MC} [TeV]" );
    hEcutNoTh2->SetYTitle( "entries (no theta2 cut)" );
    hisTreeList->Add( hEcutNoTh2 );
    hisTreeListofHistograms->Add( hEcutNoTh2 );
    
    sprintf( hname, "hEcut500" );
    hEcut500 = new TH1D( hname, "energy spectrum, after cuts", 500, fEnergyAxis_minimum_defaultValue, fEnergyAxis_maximum_defaultValue );
    hEcut500->Sumw2();
    hEcut500->SetXTitle( "energy_{MC} [TeV]" );
    hEcut500->SetYTitle( "entries" );
    hisTreeList->Add( hEcut500 );
    hisTreeListofHistograms->Add( hEcut500 );
    
    sprintf( hname, "hEcutRec" );
    hEcutRec = new TH1D( hname, "energy spectrum, after cutRecs", nbins, fEnergyAxis_minimum_defaultValue, fEnergyAxis_maximum_defaultValue );
    hEcutRec->Sumw2();
    hEcutRec->SetXTitle( "energy_{rec} [TeV]" );
    hEcutRec->SetYTitle( "entries" );
    hisTreeList->Add( hEcutRec );
    hisTreeListofHistograms->Add( hEcutRec );
    
    sprintf( hname, "hEcutRecUW" );
    hEcutRecUW = new TH1D( hname, "unweighted energy spectrum, after cutRecs", nbins, fEnergyAxis_minimum_defaultValue, fEnergyAxis_maximum_defaultValue );
    hEcutRecUW->Sumw2();
    hEcutRecUW->SetXTitle( "energy_{rec} [TeV]" );
    hEcutRecUW->SetYTitle( "entries" );
    hisTreeList->Add( hEcutRecUW );
    hisTreeListofHistograms->Add( hEcutRecUW );
    
    sprintf( hname, "hEcutRecNoTh2" );
    hEcutRecNoTh2 = new TH1D( hname, "energy spectrum, no direction cut, after cutsRecs", nbins, fEnergyAxis_minimum_defaultValue, fEnergyAxis_maximum_defaultValue );
    hEcutRecNoTh2->Sumw2();
    hEcutRecNoTh2->SetXTitle( "energy_{rec} [TeV]" );
    hEcutRecNoTh2->SetYTitle( "entries (no theta2 cut)" );
    hisTreeList->Add( hEcutRecNoTh2 );
    hisTreeListofHistograms->Add( hEcutRecNoTh2 );
    
    sprintf( hname, "gEffAreaMC" );
    gEffAreaMC = new TGraphAsymmErrors( 1 );
    gEffAreaMC->SetName( hname );
    gEffAreaMC->SetTitle( "effective area vs E_{MC}" );
    hisTreeList->Add( gEffAreaMC );
    
    sprintf( hname, "gEffAreaRec" );
    gEffAreaRec = new TGraphAsymmErrors( 1 );
    gEffAreaRec->SetName( hname );
    gEffAreaRec->SetTitle( "effective area vs E_{rec}" );
    hisTreeList->Add( gEffAreaRec );
    
    sprintf( hname, "gEffAreaNoTh2MC" );
    gEffAreaNoTh2MC = new TGraphAsymmErrors( 1 );
    gEffAreaNoTh2MC->SetName( hname );
    gEffAreaNoTh2MC->SetTitle( "effective area vs E_{MC} (no direction cut)" );
    hisTreeList->Add( gEffAreaNoTh2MC );
    
    sprintf( hname, "gEffAreaNoTh2Rec" );
    gEffAreaNoTh2Rec = new TGraphAsymmErrors( 1 );
    gEffAreaNoTh2Rec->SetName( hname );
    gEffAreaNoTh2Rec->SetTitle( "effective area vs E_{rec} (no direction cut)" );
    hisTreeList->Add( gEffAreaNoTh2Rec );
    
    // spectral weight
    sprintf( hname, "hEmcSWeight" );
    hEmcSWeight = new TProfile( hname, htitle, nbins, fEnergyAxis_minimum_defaultValue, fEnergyAxis_maximum_defaultValue, 0., 1.e12 );
    hEmcSWeight->SetXTitle( "log_{10} energy_{MC} [TeV]" );
    hEmcSWeight->SetYTitle( "spectral weight" );
    hisTreeList->Add( hEmcSWeight );
    hisTreeListofHistograms->Add( hEmcSWeight );
    
    // histograms for energy reconstruction quality
    sprintf( hname, "hEsysRec" );
    hEsysRec = new TProfile( hname, "effective area vs E_{rec}", nbins, fEnergyAxis_minimum_defaultValue, fEnergyAxis_maximum_defaultValue, -1000., 1000., "s" );
    hEsysRec->SetXTitle( "energy_{rec} [TeV]" );
    hEsysRec->SetYTitle( "log_{10} E_{rec} - log_{10} E_{MC}" );
    hisTreeList->Add( hEsysRec );
    hisTreeListofHistograms->Add( hEsysRec );
    
    sprintf( hname, "hEsysMC" );
    hEsysMC = new TProfile( hname, "energy reconstruction", nbins, fEnergyAxis_minimum_defaultValue, fEnergyAxis_maximum_defaultValue, -1000., 1000., "s" );
    hEsysMC->SetXTitle( "energy_{MC} [TeV]" );
    hEsysMC->SetYTitle( "log_{10} E_{rec} - log_{10} E_{MC}" );
    hisTreeList->Add( hEsysMC );
    hisTreeListofHistograms->Add( hEsysMC );
    
    sprintf( hname, "hEsysMCRelative" );
    hEsysMCRelative = new TProfile( hname, "energy reconstruction", nbins, fEnergyAxis_minimum_defaultValue, fEnergyAxis_maximum_defaultValue, -1000., 1000., "s" );
    hEsysMCRelative->SetXTitle( "energy_{MC} [TeV]" );
    hEsysMCRelative->SetYTitle( "energy bias (E_{rec}-E_{MC})/E_{MC}" );
    hisTreeList->Add( hEsysMCRelative );
    hisTreeListofHistograms->Add( hEsysMCRelative );
    
    sprintf( hname, "hEsysMCRelativeRMS" );
    hEsysMCRelativeRMS = new TH2D( hname, "energy reconstruction", nbins, fEnergyAxis_minimum_defaultValue, fEnergyAxis_maximum_defaultValue, 3000, -5., 5. );
    hEsysMCRelativeRMS->SetXTitle( "energy_{MC} [TeV]" );
    hEsysMCRelativeRMS->SetYTitle( "energy bias (E_{rec}-E_{MC})/E_{MC}" );
    hisTreeList->Add( hEsysMCRelativeRMS );
    hisTreeListofHistograms->Add( hEsysMCRelativeRMS );
    
    // use CTA WP Phys binning
    sprintf( hname, "hEsysMCRelative2D" );
    hEsysMCRelative2D = new TH2D( hname, "energy reconstruction", nbins, fEnergyAxis_minimum_defaultValue, fEnergyAxis_maximum_defaultValue, 300, 0., 3. );
    hEsysMCRelative2D->SetXTitle( "energy_{MC} [TeV]" );
    hEsysMCRelative2D->SetYTitle( "energy bias E_{rec}/E_{MC}" );
    hisTreeList->Add( hEsysMCRelative2D );
    hisTreeListofHistograms->Add( hEsysMCRelative2D );
    
    sprintf( hname, "hEsys2D" );
    hEsys2D = new TH2D( hname, htitle, nbins, fEnergyAxis_minimum_defaultValue, fEnergyAxis_maximum_defaultValue, 100, -0.98, 2.02 );
    hEsys2D->SetXTitle( "energy_{MC} [TeV]" );
    hEsys2D->SetYTitle( "log_{10} E_{rec} - log_{10} E_{MC}" );
    hisTreeList->Add( hEsys2D );
    hisTreeListofHistograms->Add( hEsys2D );
    
    sprintf( hname, "hResponseMatrix" );
    hResponseMatrix = new TH2D( hname, "migration matrix", nbins, fEnergyAxis_minimum_defaultValue, fEnergyAxis_maximum_defaultValue,
                                nbins, fEnergyAxis_minimum_defaultValue, fEnergyAxis_maximum_defaultValue );
    hResponseMatrix->SetYTitle( "energy_{MC} [TeV]" );
    hResponseMatrix->SetXTitle( "energy_{rec} [TeV]" );
    hisTreeList->Add( hResponseMatrix );
    hisTreeListofHistograms->Add( hResponseMatrix );
    
    sprintf( hname, "hResponseMatrixFine" );
    hResponseMatrixFine = new TH2D( hname, "migration matrix, fine binning", 500, -2.3, 2.7, 500, -2.3, 2.7 );
    hResponseMatrixFine->SetYTitle( "energy_{MC} [TeV]" );
    hResponseMatrixFine->SetXTitle( "energy_{rec} [TeV]" );
    hisTreeList->Add( hResponseMatrixFine );
    hisTreeListofHistograms->Add( hResponseMatrixFine );
    
    sprintf( hname, "hResponseMatrixQC" );
    hResponseMatrixQC = new TH2D( hname, "migration matrix, after quality cuts", nbins, fEnergyAxis_minimum_defaultValue, fEnergyAxis_maximum_defaultValue,
                                  nbins, fEnergyAxis_minimum_defaultValue, fEnergyAxis_maximum_defaultValue );
    hResponseMatrixQC->SetYTitle( "energy_{MC} [TeV]" );
    hResponseMatrixQC->SetXTitle( "energy_{rec} [TeV]" );
    hisTreeList->Add( hResponseMatrixQC );
    hisTreeListofHistograms->Add( hResponseMatrixQC );
    
    sprintf( hname, "hResponseMatrixFineQC" );
    hResponseMatrixFineQC = new TH2D( hname, "migration matrix, after quality cuts ,fine binning", 500, -2.3, 2.7, 500, -2.3, 2.7 );
    hResponseMatrixFineQC->SetYTitle( "energy_{MC} [TeV]" );
    hResponseMatrixFineQC->SetXTitle( "energy_{rec} [TeV]" );
    hisTreeList->Add( hResponseMatrixFineQC );
    hisTreeListofHistograms->Add( hResponseMatrixFineQC );
    
    // weighted rate
    // (use CTA binning, 5 bins per decade)
    sprintf( hname, "hWeightedRate" );
    hWeightedRate = new TH1D( hname, "weighted rates", 30, -2.9, 3.1 );
    hWeightedRate->Sumw2();
    hWeightedRate->SetXTitle( "energy_{rec} [TeV]" );
    hWeightedRate->SetYTitle( "entries" );
    hisTreeList->Add( hWeightedRate );
    hisTreeListofHistograms->Add( hWeightedRate );
    
    // angular resolution graphs
    for( unsigned int i = 0; i < fRunPara->fAzMin.size(); i++ )
    {
        fGraph_AngularResolution68p.push_back( 0 );
        fGraph_AngularResolution80p.push_back( 0 );
    }
    
    // individual cuts
    vector< string > iCutName;
    iCutName.push_back( "hEcutTrigger" );
    iCutName.push_back( "hEcutFiducialArea" );
    iCutName.push_back( "hEcutStereoQuality" );
    iCutName.push_back( "hEcutTelType" );
    iCutName.push_back( "hEcutDirection" );
    iCutName.push_back( "hEcutEnergyReconstruction" );
    iCutName.push_back( "hEcutGammaHadron" );
    
    for( unsigned int i = 0; i < iCutName.size(); i++ )
    {
        sprintf( hname, "h%s", iCutName[i].c_str() );
        hEcutSub.push_back( new TH1D( hname, htitle, nbins, fEnergyAxis_minimum_defaultValue, fEnergyAxis_maximum_defaultValue ) );
        hEcutSub.back()->Sumw2();
        hEcutSub.back()->SetXTitle( "energy_{MC} [TeV]" );
        hEcutSub.back()->SetYTitle( "entries" );
        hisTreeList->Add( hEcutSub.back() );
        hisTreeListofHistograms->Add( hEcutSub.back() );
    }
    
    
    // tree with results from effective area calculation
    fEffArea = new TTree( "fEffArea", "effective area values" );
    fEffArea->Branch( "ze", &ze, "ze/F" );
    fEffArea->Branch( "az", &fAzBin, "az/I" );
    fEffArea->Branch( "azMin", &fMinAz, "azMin/F" );
    fEffArea->Branch( "azMax", &fMaxAz, "azMax/F" );
    if( !fRunPara->fEffArea_short_writing )
    {
        fEffArea->Branch( "Xoff", &fXoff, "Xoff/F" );
        fEffArea->Branch( "Yoff", &fYoff, "Yoff/F" );
    }
    fEffArea->Branch( "Woff", &fWoff, "Woff/F" );
    fEffArea->Branch( "noise", &fTNoise, "noise/I" );
    if( !fRunPara->fEffArea_short_writing )
    {
        fEffArea->Branch( "noisePE", &fTNoisePE, "noisePE/F" );
    }
    fEffArea->Branch( "pedvar", &fTPedvar, "pedvar/F" );
    fEffArea->Branch( "index", &fSpectralIndex, "index/F" );
    fEffArea->Branch( "nbins", &nbins, "nbins/I" );
    fEffArea->Branch( "e0", e0, "e0[nbins]/F" );  // log10( energy ) in [TeV]
    fEffArea->Branch( "eff", eff, "eff[nbins]/F" ); // effective area vs MC energy
    fEffArea->Branch( "eff_error", eff_error, "eff_error[nbins]/F" );
    fEffArea->Branch( "esys_rel", esys_rel, "esys_rel[nbins]/F" );
    if( !fRunPara->fEffArea_short_writing )
    {
        fEffArea->Branch( "seff_L", seff_L, "seff_L[nbins]/F" );
        fEffArea->Branch( "seff_U", seff_U, "seff_U[nbins]/F" );
    }
    fEffArea->Branch( "Rec_nbins", &Rec_nbins, "Rec_nbins/I" );
    fEffArea->Branch( "Rec_e0", Rec_e0, "Rec_e0[Rec_nbins]/F" ); // log10( energy ) in [TeV]
    fEffArea->Branch( "Rec_eff", Rec_eff, "Rec_eff[Rec_nbins]/F" ); // effective area vs reconstructed energy (approximation)
    fEffArea->Branch( "Rec_eff_error", Rec_eff_error, "Rec_eff_error[nbins]/F" );
    if( !fRunPara->fEffArea_short_writing )
    {
        fEffArea->Branch( "Rec_seff_L", Rec_seff_L, "Rec_seff_L[Rec_nbins]/F" );
        fEffArea->Branch( "Rec_seff_U", Rec_seff_U, "Rec_seff_U[Rec_nbins]/F" );
    }
    fEffArea->Branch( "Rec_angRes_p68", Rec_angRes_p68, "Rec_angRes_p68[Rec_nbins]/F" );
    fEffArea->Branch( "Rec_angRes_p80", Rec_angRes_p80, "Rec_angRes_p80[Rec_nbins]/F" );
    if( !fRunPara->fEffArea_short_writing )
    {
        fEffArea->Branch( hisTreeList, 64000, 1 );
    }
    fEffArea->SetMarkerStyle( 20 );
    
    fAcceptance_AfterCuts_tree = new TTree( "Acceptance_AfterCuts" , "Info to construct background map" );
    fAcceptance_AfterCuts_tree->Branch( "Xoff_aC", &fXoff_aC, "Xoff_aC/F" );
    fAcceptance_AfterCuts_tree->Branch( "Yoff_aC", &fYoff_aC, "Yoff_aC/F" );
    fAcceptance_AfterCuts_tree->Branch( "Xoff_derot_aC", &fXoff_derot_aC, "Xoff_derot_aC/F" );
    fAcceptance_AfterCuts_tree->Branch( "Yoff_derot_aC", &fYoff_derot_aC, "Yoff_derot_aC/F" );
    fAcceptance_AfterCuts_tree->Branch( "Erec", &fErec, "Erec/F" );
    fAcceptance_AfterCuts_tree->Branch( "EMC", &fEMC, "EMC/F" );
    fAcceptance_AfterCuts_tree->Branch( "CRweight", &fCRweight, "CRweight/F" );
    
    fsolid_angle_norm_done = false;
    fsolid_angle_norm = 1.;
    
}

vector< TH1D* > VEffectiveAreaCalculator::initializeHistogramsVectorH1D( TH1D* h, string iName, unsigned int i )
{
    char hname[400];
    vector< TH1D* > iT_TH1D;
    
    for( unsigned int j = 0; j < fVMinAz.size(); j++ )
    {
        sprintf( hname, "%s_%d_%d", iName.c_str(), i, j );
        if( h )
        {
            iT_TH1D.push_back( ( TH1D* )h->Clone( hname ) );
            if( hisVList )
            {
                hisVList->Add( iT_TH1D.back() );
            }
        }
        else
        {
            iT_TH1D.push_back( 0 );
        }
    }
    return iT_TH1D;
}

vector< TH2D* > VEffectiveAreaCalculator::initializeHistogramsVectorH2D( TH2D* h, string iName, unsigned int i )
{
    char hname[400];
    vector< TH2D* > iT_TH2D;
    
    for( unsigned int j = 0; j < fVMinAz.size(); j++ )
    {
        sprintf( hname, "%s_%d_%d", iName.c_str(), i, j );
        if( h )
        {
            iT_TH2D.push_back( ( TH2D* )h->Clone( hname ) );
            if( hisVList )
            {
                hisVList->Add( iT_TH2D.back() );
            }
        }
        else
        {
            iT_TH2D.push_back( 0 );
        }
    }
    return iT_TH2D;
}

vector< TProfile* > VEffectiveAreaCalculator::initializeHistogramsVectorHProfile( TProfile* h, string iName, unsigned int i )
{
    char hname[400];
    vector< TProfile* > iT_TProfile;
    
    for( unsigned int j = 0; j < fVMinAz.size(); j++ )
    {
        sprintf( hname, "%s_%d_%d", iName.c_str(), i, j );
        if( h )
        {
            iT_TProfile.push_back( ( TProfile* )h->Clone( hname ) );
            if( hisVList )
            {
                hisVList->Add( iT_TProfile.back() );
            }
        }
        else
        {
            iT_TProfile.push_back( 0 );
        }
    }
    return iT_TProfile;
}



/*

   set up vectors with histograms for azimuth and spectral index bins

*/
void VEffectiveAreaCalculator::initializeHistograms( vector< double > iAzMin, vector< double > iAzMax, vector< double > iSpectralIndex )
{
    fVMinAz = iAzMin;
    fVMaxAz = iAzMax;
    fVSpectralIndex = iSpectralIndex;
    
    char hname[200];
    
    // temporary histograms for effective area calculation
    for( unsigned int i = 0; i < fVSpectralIndex.size(); i++ )
    {
        hVEmc.push_back( initializeHistogramsVectorH1D( hEmc, "hVEmc", i ) );
        hVEcut.push_back( initializeHistogramsVectorH1D( hEcut, "hVEcut", i ) );
        hVEcutUW.push_back( initializeHistogramsVectorH1D( hEcutUW, "hVEcutUW", i ) );
        hVEcut500.push_back( initializeHistogramsVectorH1D( hEcut500, "hVEcut500", i ) );
        hVEcutRec.push_back( initializeHistogramsVectorH1D( hEcutRec, "hVEcutRec", i ) );
        hVEcutRecUW.push_back( initializeHistogramsVectorH1D( hEcutRecUW, "hVEcutRecUW", i ) );
        hVEcutNoTh2.push_back( initializeHistogramsVectorH1D( hEcutNoTh2, "hVEcutNoTh2", i ) );
        hVEcutRecNoTh2.push_back( initializeHistogramsVectorH1D( hEcutRecNoTh2, "hVEcutRecNoTh2", i ) );
        hVEmcSWeight.push_back( initializeHistogramsVectorHProfile( hEmcSWeight, "hVEmcSWeight", i ) );
        hVEsysRec.push_back( initializeHistogramsVectorHProfile( hEsysRec, "hVEsysRec", i ) );
        hVEsysMC.push_back( initializeHistogramsVectorHProfile( hEsysMC, "hVEsysMC", i ) );
        hVEsysMCRelative.push_back( initializeHistogramsVectorHProfile( hEsysMCRelative, "hVEsysMCRelative", i ) );
        hVEsysMCRelativeRMS.push_back( initializeHistogramsVectorH2D( hEsysMCRelativeRMS, "hVEsysMCRelativeRMS", i ) );
        hVEsysMCRelative2D.push_back( initializeHistogramsVectorH2D( hEsysMCRelative2D, "hVEsysMCRelative2D", i ) );
        hVEsys2D.push_back( initializeHistogramsVectorH2D( hEsys2D, "hVEsys2D", i ) );
        hVResponseMatrixFine.push_back( initializeHistogramsVectorH2D( hResponseMatrixFine, "hVResponseMatrixFine", i ) );
        hVResponseMatrixFineQC.push_back( initializeHistogramsVectorH2D( hResponseMatrixFineQC, "hVResponseMatrixFineQC", i ) );
        hVResponseMatrix.push_back( initializeHistogramsVectorH2D( hResponseMatrix, "hVResponseMatrix", i ) );
        hVResponseMatrixQC.push_back( initializeHistogramsVectorH2D( hResponseMatrixQC, "hVResponseMatrixQC", i ) );
        hVWeightedRate.push_back( initializeHistogramsVectorH1D( hWeightedRate, "hVWeightedRate", i ) );
        vector< vector< TH1D* > > i_temp;
        for( unsigned int e = 0; e < hEcutSub.size(); e++ )
        {
            sprintf( hname, "hV%s", hEcutSub[e]->GetName() );
            i_temp.push_back( initializeHistogramsVectorH1D( hEcutSub[e], hname, i ) );
        }
        hVEcutSub.push_back( i_temp );
        
    }
}


/*!
 *
 *  CALLED TO USE (READ) EFFECTIVE AREAS
 *
 *  this constructor is called to GET the effective area from a file and calculate energy spectra
 *
 *  called from anasum
 *
 */
VEffectiveAreaCalculator::VEffectiveAreaCalculator( string iInputFile, double azmin, double azmax, double ipedvar,
        double iSpectralIndex, vector< double > iMCZe,
        int iSmoothIter, double iSmoothThreshold, int iEffectiveAreaVsEnergyMC )
{
    reset();
    
    fRunPara = 0;
    
    // MC intervalls
    fMCZe = iMCZe;
    
    // no effective area file present
    bNOFILE = false;
    // no input file available, return always 1 for all corrections
    if( iInputFile == "NOFILE" )
    {
        bNOFILE = true;
        return;
    }
    // effective area smoothing
    fSmoothIter      = iSmoothIter;
    fSmoothThreshold = iSmoothThreshold;
    
    // no weighting
    fSpectralWeight = 0;
    
    // effective areas vs E_MC or E_rec
    fEffectiveAreaVsEnergyMC = iEffectiveAreaVsEnergyMC;
    
    // mean effective area
    gMeanEffectiveArea = new TGraphAsymmErrors( 1 );
    gMeanEffectiveArea->SetName( "gMeanEffectiveArea" );
    gMeanEffectiveArea->SetMarkerStyle( 20 );
    gMeanEffectiveArea->SetMarkerColor( fEffectiveAreaVsEnergyMC + 1 );
    gMeanEffectiveArea->SetLineColor( fEffectiveAreaVsEnergyMC + 1 );
    
    // mean effective area for Time BINS
    
    gTimeBinnedMeanEffectiveArea = new TGraph2DErrors( 1 );
    gTimeBinnedMeanEffectiveArea->SetName( "gTimeBinnedMeanEffectiveArea" );
    gTimeBinnedMeanEffectiveArea->SetMarkerStyle( 20 );
    gTimeBinnedMeanEffectiveArea->SetMarkerColor( fEffectiveAreaVsEnergyMC + 1 );
    gTimeBinnedMeanEffectiveArea->SetLineColor( fEffectiveAreaVsEnergyMC + 1 );
    
    for( int i = 0; i < gTimeBinnedMeanEffectiveArea->GetN(); i++ )
    {
        gTimeBinnedMeanEffectiveArea->SetPoint( i, 0., 0., 0. );
        gTimeBinnedMeanEffectiveArea->SetPointError( i, 0., 0., 0. );
    }
    
    
    // current directory
    fGDirectory = gDirectory;
    
    // test if input file with effective areas exists
    if( iInputFile.find( "IGNOREEFFECTIVEAREA" ) == string::npos )
    {
        iInputFile = VUtilities::testFileLocation( iInputFile, "EffectiveAreas", true );
    }
    if( iInputFile.size() == 0 )
    {
        cout << "VEffectiveAreaCalculator::VEffectiveAreaCalculator error: input file string of zero length" << endl;
        cout << "exiting..." << endl;
        exit( EXIT_FAILURE );
    }
    else if( iInputFile.find( "IGNOREEFFECTIVEAREA" ) != string::npos )
    {
       cout << "ignoring effective areas: ";
       cout << "all energy spectra will be invalid" << endl;
       bNOFILE = true;
    }
    else
    {
        TFile fIn( iInputFile.c_str() );
        if( fIn.IsZombie() )
        {
            cout << "VEffectiveAreaCalculator::VEffectiveAreaCalculator error opening file with effective areas: " << iInputFile << endl;
            cout << "exiting..." << endl;
            exit( EXIT_FAILURE );
        }
        cout << "\t reading effective areas from " << fIn.GetName() << endl;
        
        if( !initializeEffectiveAreasFromHistograms( ( TTree* )gDirectory->Get( "fEffArea" ), ( TH1D* )gDirectory->Get( "hEmc" ), azmin, azmax, iSpectralIndex, ipedvar ) )
        {
            cout << "VEffectiveAreaCalculator ERROR: no effective areas found" << endl;
            cout << "all energy spectra will be invalid" << endl;
            bNOFILE = true;
        }
        fIn.Close();
        
        if( fGDirectory )
        {
            fGDirectory->cd();
        }
    }
}


float VEffectiveAreaCalculator::getAzMean( float azmin, float azmax )
{
    // mean azimuth angle
    float iAzMean = 0.;
    if( azmin > 120. && azmax < -120. )
    {
        azmax += 360.;
    }
    else if( azmin < -150. && azmax > 120. )
    {
        azmin += 360.;
    }
    
    iAzMean = 0.5 * ( azmin + azmax );
    if( iAzMean > 180. )
    {
        iAzMean -= 360.;
    }
    
    return iAzMean;
}

/*
 * interpolate between two vectors
 *
 * (with or without 1/cos)
 */
vector< float > VEffectiveAreaCalculator::interpolate_effectiveArea( double iV, double iVLower, double iVupper,
        vector< float > iElower, vector< float > iEupper, bool iCos )
{
    vector< float > i_temp;
    if( iElower.size() == iEupper.size() )
    {
        i_temp.assign( iElower.size(), 0. );
        for( unsigned int i = 0; i < iElower.size(); i++ )
        {
            i_temp[i] = VStatistics::interpolate( iElower[i], iVLower, iEupper[i], iVupper, iV, iCos, 0.5, -90. );
        }
        return i_temp;
    }
    
    return i_temp;
}


/*
 *
 *  CALLED TO USE (READ) EFFECTIVE AREAS
 *
 *  called from anasum
 *
 */
bool VEffectiveAreaCalculator::initializeEffectiveAreasFromHistograms( TTree* iEffArea, TH1D* i_hEMC,
        double azmin, double azmax,
        double iSpectralIndex, double ipedvar )
{
    if( !iEffArea )
    {
        return false;
    }
    if( !i_hEMC )
    {
        cout << "----- Warning -----" << endl;
        cout << "  no MC histogram found to determine energy binning " << endl;
        cout << "   assume default binning: ";
        cout << "   " << fEnergyAxis_minimum_defaultValue << " " << fEnergyAxis_maximum_defaultValue;
        cout << " " << nbins;
        cout << endl;
        cout << "---- End of Warning ----" << endl;
        i_hEMC = new TH1D( "hEmc", "", nbins, fEnergyAxis_minimum_defaultValue, fEnergyAxis_maximum_defaultValue );
    }
    if( iEffArea->GetEntries() == 0 )
    {
        cout << "VEffectiveAreaCalculator::initializeEffectiveAreasFromHistograms: empty effective area tree" << endl;
        return false;
    }
    
    // mean azimuth angle
    float iAzMean = getAzMean( azmin, azmax );
    
    ////////////////////////////
    // define input tree
    float TazMin = 0.;
    float TazMax = 0.;
    float Tpedvar = 1.;
    int   nbins_MC = 0;
    float e0_MC[1000];
    iEffArea->SetBranchAddress( "azMin", &TazMin );
    iEffArea->SetBranchAddress( "azMax", &TazMax );
    iEffArea->SetBranchAddress( "pedvar", &Tpedvar );
    iEffArea->SetBranchAddress( "index", &fSpectralIndex );
    iEffArea->SetBranchAddress( "ze", &ze );
    iEffArea->SetBranchAddress( "Woff", &fWoff );
    // MC binning required for systematic error estimation
    iEffArea->SetBranchAddress( "nbins", &nbins_MC );
    iEffArea->SetBranchAddress( "e0", e0_MC );
    // effective areas vs true energy
    if( fEffectiveAreaVsEnergyMC == 0 )
    {
        iEffArea->SetBranchAddress( "nbins", &nbins );
        iEffArea->SetBranchAddress( "e0", e0 );
        iEffArea->SetBranchAddress( "eff", eff );
    }
    // effective area vs reconstruction energy
    // this method should be used for the correction unfolding method
    else if( fEffectiveAreaVsEnergyMC == 1 )
    {
        iEffArea->SetBranchAddress( "Rec_nbins", &nbins );
        iEffArea->SetBranchAddress( "Rec_e0", e0 );
        iEffArea->SetBranchAddress( "Rec_eff", eff );
    }
    // bias in energy reconstruction
    iEffArea->SetBranchAddress( "esys_rel", esys_rel );
    
    ////////////////////////////////////////////////////////////////////////////////////
    // prepare the energy vectors
    // (binning should be the same for all entries in the effective area tree)
    ////////////////////////////////////////////////////////////////////////////////////
    
    iEffArea->GetEntry( 0 );
    
    if( !i_hEMC )
    {
        cout << "VEffectiveAreaCalculator::initializeEffectiveAreasFromHistograms error: no effective area histogram found" << endl;
        return false;
    }
    
    fEff_E0.clear();
    for( int b = 1; b <= i_hEMC->GetNbinsX(); b++ )
    {
        fEff_E0.push_back( i_hEMC->GetBinCenter( b ) );
    }
    
    fVTimeBinnedMeanEffectiveArea.assign( i_hEMC->GetNbinsX(), 0. );
    // temporary vectors filled into the effective area maps later
    // effective areas (energy axis depend on fEffectiveAreaVsEnergyMC)
    vector< float > i_temp_Eff( i_hEMC->GetNbinsX(), 0. );
    // bias in energy reconstruction (energy axis is in E_true)
    vector< float > i_temp_Esys( i_hEMC->GetNbinsX(), 0. );
    
    cout << "\t selecting effective areas for mean az " << iAzMean << " deg, spectral index ";
    cout << iSpectralIndex << ", noise level " << ipedvar << endl;
    cout << "\t\ttotal number of curves: " << iEffArea->GetEntries();
    cout << ", total number of bins on energy axis: " << fEff_E0.size() << endl;
    
    fNTimeBinnedMeanEffectiveArea = 0;
    if( gTimeBinnedMeanEffectiveArea )
    {
        for( int i = 0; i < gTimeBinnedMeanEffectiveArea->GetN(); i++ )
        {
            gTimeBinnedMeanEffectiveArea->SetPoint( i, 0., 0., 0. );
            gTimeBinnedMeanEffectiveArea->SetPointError( i, 0., 0., 0. );
        }
    }
    
    /////////////////////////////////////////////////////////////////////////////////////
    /////////////////////////////////////////////////////////////////////////////////////
    // fill ze, woff, noise, index vectors
    /////////////////////////////////////////////////////////////////////////////////////
    /////////////////////////////////////////////////////////////////////////////////////
    bool i_ze_F = false;
    unsigned int i_index_ze = 0;
    bool i_woff_F = false;
    unsigned int i_index_woff = 0;
    bool i_noise_F = false;
    unsigned int i_index_noise = 0;
    bool i_index_F = false;
    unsigned int i_index_index = 0;
    int iIndexAz = 0;
    double iInvMean = 0.;
    double iInvMax = 1.e5;
    
    bool fLotsOfPrintOuts = false;
    
    ///////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////////
    // loop over all entries in effective area tree
    // (not sure if this is really necessary, in the end a few entries are only needed)
    for( int i = 0; i < iEffArea->GetEntries(); i++ )
    {
        iEffArea->GetEntry( i );
        
        // check binning of effective areas
        if( i_hEMC && ( int )i_hEMC->GetNbinsX() != ( int )fEff_E0.size() )
        {
            cout << "VEffectiveAreaCalculator::initializeEffectiveAreasFromHistograms error: effective area curve with different binning";
            cout << " " << i_hEMC->GetNbinsX() << " " << fEff_E0.size() << endl;
            cout << "abort reading effective area tree..." << endl;
            return false;
        }
        ///////////////////////////////////////////////////
        // check the azimuth range
        ///////////////////////////////////////////////////
        
        // expect a couple of az bins and then a last bin with the average azimuth bin; typically [-1000., 1000.]
        // this is the sign to read effective areas
        if( fabs( TazMin ) > 5.e2 || fabs( TazMax ) > 5.e2 )
        {
            iInvMax = 1.e5;
            iEffArea->GetEntry( iIndexAz );
            
            ///////////////////////////////////////////////////
            // zenith angle
            ///////////////////////////////////////////////////
            i_ze_F = false;
            for( unsigned z = 0; z < fZe.size(); z++ )
            {
                // this has to be a relatively large value due to wobble offsets up to 2.0 deg
                if( fabs( fZe[z] - ze ) < 2.0 )
                {
                    i_index_ze = z;
                    i_ze_F = true;
                    break;
                }
            }
            if( !i_ze_F )
            {
                for( unsigned int w = 0; w < fMCZe.size(); w++ )
                {
                    if( fabs( fMCZe[w] - ze ) < 2.0 )
                    {
                        fZe.push_back( fMCZe[w] );
                        break;
                    }
                }
                i_index_ze = fZe.size() - 1;
            }
            ///////////////////////////////////////////////////
            // wobble offset
            ///////////////////////////////////////////////////
            if( i_index_ze < fEff_WobbleOffsets.size() )
            {
                i_woff_F = false;
                for( unsigned z = 0; z < fEff_WobbleOffsets[i_index_ze].size(); z++ )
                {
                    if( fabs( fEff_WobbleOffsets[i_index_ze][z] - fWoff ) < 0.05 )
                    {
                        i_index_woff = z;
                        i_woff_F = true;
                        break;
                    }
                }
                if( !i_woff_F )
                {
                    fEff_WobbleOffsets[i_index_ze].push_back( fWoff );
                    i_index_woff = fEff_WobbleOffsets[i_index_ze].size() - 1;
                }
            }
            else
            {
                vector< double > itemp;
                itemp.push_back( fWoff );
                fEff_WobbleOffsets.push_back( itemp );
                i_index_woff = fEff_WobbleOffsets[i_index_ze].size() - 1;
            }
            ///////////////////////////////////////////////////
            // noise level
            ///////////////////////////////////////////////////
            if( i_index_ze < fEff_Noise.size() )
            {
                if( i_index_woff < fEff_Noise[i_index_ze].size() )
                {
                    i_noise_F = false;
                    for( unsigned w = 0; w < fEff_Noise[i_index_ze][i_index_woff].size(); w++ )
                    {
                        if( fabs( fEff_Noise[i_index_ze][i_index_woff][w] - Tpedvar ) < 0.005 )
                        {
                            i_index_noise = w;
                            i_noise_F = true;
                            break;
                        }
                    }
                    if( !i_noise_F )
                    {
                        fEff_Noise[i_index_ze][i_index_woff].push_back( Tpedvar );
                        i_index_noise = fEff_Noise[i_index_ze][i_index_woff].size() - 1;
                    }
                }
                else
                {
                    vector< double > itemp;
                    itemp.push_back( Tpedvar );
                    fEff_Noise[i_index_ze].push_back( itemp );
                    i_index_noise = fEff_Noise[i_index_ze][i_index_woff].size() - 1;
                }
            }
            else
            {
                vector< double > itemp;
                itemp.push_back( Tpedvar );
                vector< vector< double > > iitemp;
                iitemp.push_back( itemp );
                fEff_Noise.push_back( iitemp );
                i_index_noise = fEff_Noise[i_index_ze][i_index_woff].size() - 1;
            }
            ///////////////////////////////////////////////////
            // spectral index
            ///////////////////////////////////////////////////
            if( i_index_ze < fEff_SpectralIndex.size() )
            {
                if( i_index_woff < fEff_SpectralIndex[i_index_ze].size() )
                {
                    if( i_index_noise < fEff_SpectralIndex[i_index_ze][i_index_woff].size() )
                    {
                        i_index_F = false;
                        for( unsigned s = 0; s < fEff_SpectralIndex[i_index_ze][i_index_woff][i_index_noise].size(); s++ )
                        {
                            if( fabs( fEff_SpectralIndex[i_index_ze][i_index_woff][i_index_noise][s] - fSpectralIndex ) < 0.005 )
                            {
                                i_index_index = s;
                                i_index_F = true;
                                break;
                            }
                        }
                        if( !i_index_F )
                        {
                            fEff_SpectralIndex[i_index_ze][i_index_woff][i_index_noise].push_back( fSpectralIndex );
                            i_index_index = fEff_SpectralIndex[i_index_ze][i_index_woff][i_index_noise].size() - 1;
                        }
                    }
                    else
                    {
                        vector< double > itemp;
                        itemp.push_back( fSpectralIndex );
                        fEff_SpectralIndex[i_index_ze][i_index_woff].push_back( itemp );
                        i_index_index = fEff_SpectralIndex[i_index_ze][i_index_woff][i_index_noise].size() - 1;
                    }
                }
                else
                {
                    vector< double > itemp;
                    itemp.push_back( fSpectralIndex );
                    vector< vector< double > > iitemp;
                    iitemp.push_back( itemp );
                    fEff_SpectralIndex[i_index_ze].push_back( iitemp );
                    i_index_index = fEff_SpectralIndex[i_index_ze][i_index_woff][i_index_noise].size() - 1;
                }
            }
            else
            {
                vector< double > itemp;
                itemp.push_back( fSpectralIndex );
                vector< vector< double > > iitemp;
                iitemp.push_back( itemp );
                vector< vector< vector< double > > > iiitemp;
                iiitemp.push_back( iitemp );
                fEff_SpectralIndex.push_back( iiitemp );
                i_index_index = fEff_SpectralIndex[i_index_ze][i_index_woff][i_index_noise].size() - 1;
            }
            unsigned int i_ID = i_index_index + 100 * ( i_index_noise + 100 * ( i_index_woff + 100 * i_index_ze ) );
            if( fLotsOfPrintOuts )
            {
                cout << i_index_ze << " " << i_index_woff << " " << i_index_noise << " " << i_index_index << "\t" << i_ID << endl;
            }
            
            ///////////////////////////////////////////////////
            // read effective area and load into maps
            ///////////////////////////////////////////////////
            if( nbins <= 0 )
            {
                cout << "WARNING : incomplete effective areas for id " << i_ID << endl;
                cout << i_index_index << "\t" << i_index_noise << "\t" << i_index_woff << "\t" << i_index_ze << endl;
                cout << "in bool VEffectiveAreaCalculator::initializeEffectiveAreasFromHistograms(";
                cout << "TTree *iEffArea, double azmin, double azmax, double iSpectralIndex )" << endl;
                cout << "Missing effective area:";
                if( i_index_ze < fZe.size() )
                {
                    cout << " ze = " << fZe[i_index_ze] << " [deg],";
                }
                if( i_index_ze < fEff_WobbleOffsets.size() && i_index_woff < fEff_WobbleOffsets[i_index_ze].size() )
                {
                    cout << " woff = " << fEff_WobbleOffsets[i_index_ze][i_index_woff] << " [deg]";
                }
                if( i_index_ze < fEff_Noise.size() &&
                        i_index_woff < fEff_Noise[i_index_ze].size() &&
                        i_index_noise < fEff_Noise[i_index_ze][i_index_woff].size() )
                {
                    cout << " noise = " << fEff_Noise[i_index_ze][i_index_woff][i_index_noise];
                }
                if( i_index_ze < fEff_SpectralIndex.size() && i_index_woff < fEff_SpectralIndex[i_index_ze].size() &&
                        i_index_noise < fEff_SpectralIndex[i_index_ze][i_index_woff].size() &&
                        i_index_index < fEff_SpectralIndex[i_index_ze][i_index_woff][i_index_noise].size() )
                {
                    cout << " spectral index = " << fEff_SpectralIndex[i_index_ze][i_index_woff][i_index_noise][i_index_index];
                }
                cout << endl;
                cout << "please check if this falls into your parameter space" << endl;
            }
            // effective areas vs energy (decision on vs MC, rec energy etc. has been made earlier)
            for( unsigned int e = 0; e < fEff_E0.size(); e++ )
            {
                i_temp_Eff[e] = 0.;
                for( int j = 0; j < nbins; j++ )
                {
                    if( TMath::Abs( e0[j] - fEff_E0[e] ) < 1.e-5 )
                    {
                        i_temp_Eff[e] = eff[j];
                    }
                }
                i_temp_Esys[e] = 0.;
                for( int j = 0; j < nbins_MC; j++ )
                {
                    if( TMath::Abs( e0_MC[j] - fEff_E0[e] ) < 1.e-5 )
                    {
                        i_temp_Esys[e]  = esys_rel[j];
                    }
                    
                }
            }
            fEffArea_map[i_ID]         = i_temp_Eff;
            fEff_EsysMCRelative[i_ID]  = i_temp_Esys;
            
            // this is neeeded only if there are no azimuth dependent effective areas
            iIndexAz++;
            
            continue;
        }
        ///////////////////////////////////////////////////
        // test az bin
        ///////////////////////////////////////////////////
        // expect bin like [135,-135]
        iInvMean = getAzMean( TazMin, TazMax );
        if( TazMax < 0. && TazMin > 0. )
        {
            double iT =  fabs( iAzMean - iInvMean );
            if( iT > 180. )
            {
                iT = fabs( iT - 360. );
            }
            if( iT < iInvMax )
            {
                iInvMax = iT;
                iIndexAz = i;
            }
            if( iAzMean < 0. && iAzMean > TazMax )
            {
                continue;
            }
            else if( iAzMean > 0. && iAzMean < TazMin )
            {
                continue;
            }
        }
        else
        {
            if( fabs( iAzMean - iInvMean ) < iInvMax )
            {
                iInvMax = fabs( iAzMean - iInvMean );
                iIndexAz = i;
            }
        }
    }
    ///////////////////////////////////////////////////
    if( fZe.size() == 0 )
    {
        cout << "VEffectiveAreaCalculator::initializeEffectiveAreasFromHistograms error: no effective areas found in effective area tree" << endl;
        cout << "exiting..." << endl;
        exit( EXIT_FAILURE );
    }
    cout << "\t use histograms for effective areas (number of zenith angle intervals: " << fZe.size() << "; ";
    for( unsigned int i = 0; i < fZe.size() - 1; i++ )
    {
        cout << fZe[i] << ", ";
    }
    if( fZe.size() > 0 )
    {
        cout << fZe[fZe.size() - 1];
    }
    cout << " deg)" << endl;
    if( fEffectiveAreaVsEnergyMC == 0 )
    {
        cout << "\t (effective area vs MC energy)" << endl;
    }
    else
    {
        cout << "\t (effective area vs reconstructed energy)" << endl;
    }
    
    if( fSmoothIter > 0 )
    {
        smoothEffectiveAreas( fEffArea_map );
    }
    
    ///////////////////////////////////////////////////
    if( fLotsOfPrintOuts )
    {
        cout << "ze size " << fZe.size() << endl;
        for( unsigned int z = 0; z < fZe.size(); z++ )
        {
            cout << "ze " << z << " " << fZe[z] << endl;
            if( z < fEff_WobbleOffsets.size() )
            {
                cout << "\t woff size " << fEff_WobbleOffsets[z].size() << endl;
                for( unsigned int w = 0; w < fEff_WobbleOffsets[z].size(); w++ )
                {
                    cout << "\t woff " << w << " " << fEff_WobbleOffsets[z][w] << endl;
                    if( z < fEff_Noise.size() && w < fEff_Noise[z].size() )
                    {
                        cout << "\t\t noise size " << fEff_Noise[z][w].size() << endl;
                        for( unsigned n = 0; n < fEff_Noise[z][w].size(); n++ )
                        {
                            cout << "\t\t noise " << n << " " << fEff_Noise[z][w][n] << endl;
                            if( z < fEff_SpectralIndex.size() && w < fEff_SpectralIndex[z].size() && n < fEff_SpectralIndex[z][w].size() )
                            {
                                cout << "\t\t\t index size " << fEff_SpectralIndex[z][w][n].size() << endl;
                                for( unsigned s = 0; s < fEff_SpectralIndex[z][w][n].size(); s++ )
                                {
                                    cout << "\t\t\t index " << s << " " << fEff_SpectralIndex[z][w][n][s] << endl;
                                    unsigned int i_ID = s + 100 * ( n + 100 * ( w + 100 * z ) );
                                    if( fEffArea_map.find( i_ID ) != fEffArea_map.end() )
                                    {
                                        cout << "\t\t\t map " << fEffArea_map[i_ID].size() << endl;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    ///////////////////////////////////////////////////
    
    return true;
}


/////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////

VEffectiveAreaCalculator::~VEffectiveAreaCalculator()
{
    if( gMeanEffectiveArea )
    {
        delete gMeanEffectiveArea;
    }
    if( gTimeBinnedMeanEffectiveArea )
    {
        delete gTimeBinnedMeanEffectiveArea;
    }
    if( gMeanSystematicErrorGraph )
    {
        delete gMeanSystematicErrorGraph;
    }
}


void VEffectiveAreaCalculator::reset()
{
    gMeanEffectiveArea = 0;
    gTimeBinnedMeanEffectiveArea = 0;
    
    fMC_ScatterArea = 0.;
    
    bNOFILE = true;
    fGDirectory = 0;
    
    fSpectralIndex = 2.0;
    
    // Important: changing this means probably that the values used in
    // VEffectiveAreaCalculatorMCHistograms have to be changed as well
    fEnergyAxis_minimum_defaultValue = -2.0;
    fEnergyAxis_maximum_defaultValue = 4.0;
    
    fCuts = 0;
    
    fTNoise = 0;
    fTNoisePE = 0.;
    fTPedvar = 0.;
    
    fAzBin = 0;
    fMinAz = -1.e3;
    fMaxAz = 1.e3;
    fEffectiveAreaVsEnergyMC = 1;
    
    gEffAreaMC = 0;
    gEffAreaRec = 0;
    
    gEffAreaNoTh2MC = 0;
    gEffAreaNoTh2Rec = 0;
    
    hEmc = 0;
    hEcut = 0;
    hEcutUW = 0;
    hEcutRec = 0;
    hEcutRecUW = 0;
    hEcut500 = 0;
    hEcutNoTh2 = 0;
    hEcutRecNoTh2 = 0;
    hEmcSWeight = 0;
    hEsysRec = 0;
    hEsysMC = 0;
    hEsysMCRelative = 0;
    hEsysMCRelative2D = 0;
    hEsys2D = 0;
    hResponseMatrix = 0;
    hResponseMatrixQC = 0;
    hResponseMatrixFine = 0;
    hResponseMatrixFineQC = 0;
    fEffArea = 0;
    hisTreeList = 0;
    hisTreeListofHistograms = 0;
    hisVList = 0;
    
    fEffArea = 0;
    ze = 0.;
    nbins = 60;
    Rec_nbins = 0;
    for( int i = 0; i < 1000; i++ )
    {
        e0[i] = 0.;
        eff[i] = 0.;
        eff_error[i] = 0.;
        esys_rel[i] = 0.;
        seff_L[i] = 0.;
        seff_U[i] = 0.;
        Rec_e0[i] = 0.;
        Rec_eff[i] = 0.;
        Rec_eff_error[i] = 0.;
        Rec_seff_L[i] = 0.;
        Rec_seff_U[i] = 0.;
        Rec_angRes_p68[i] = 0.;
        Rec_angRes_p80[i] = 0.;
    }
    
    fEffectiveAreas_meanZe = 0.;
    fEffectiveAreas_meanWoff = 0.;
    fEffectiveAreas_meanPedVar = 0.;
    fEffectiveAreas_meanIndex = 0.;
    fEffectiveAreas_meanN = 0.;
    
    gMeanSystematicErrorGraph = 0;
    
    fXoff_aC = -99;
    fYoff_aC = -99;
    fXoff_derot_aC = -99;
    fYoff_derot_aC = -99;
    fErec = -99;
    fEMC = -99;
    fCRweight = -99;
    
    
}

double VEffectiveAreaCalculator::getMCSolidAngleNormalization()
{
    double iSolAngleNorm = 1.;
    if( fCuts && fRunPara )
    {
        if( fRunPara->fViewcone_max > 0. && fCuts->fCut_CameraFiducialSize_MC_max > 0. && fCuts->fCut_CameraFiducialSize_MC_max < 1000.
                && fCuts->fCut_CameraFiducialSize_MC_max < fRunPara->fViewcone_max )
        {
            // solid angle of simulated showers
            double iSN_mc = ( 1. - cos( fRunPara->fViewcone_max * TMath::DegToRad() ) );
            if( fRunPara->fViewcone_min > 0. )
            {
                iSN_mc -= ( 1. - cos( fRunPara->fViewcone_min * TMath::DegToRad() ) );
            }
            // solid angle of angular bin
            double iSN_cu = ( 1. - cos( fCuts->fCut_CameraFiducialSize_MC_max * TMath::DegToRad() ) );
            if( fCuts->fCut_CameraFiducialSize_MC_min > 0. )
            {
                iSN_cu -= ( 1. - cos( fCuts->fCut_CameraFiducialSize_MC_min * TMath::DegToRad() ) );
            }
            
            if( iSN_mc > 0. )
            {
                iSolAngleNorm = iSN_cu / iSN_mc;
            }
        }
    }
    
    return iSolAngleNorm;
}

/*

   read spectra of MC events from mscw file (filled in eventdisplay)

*/
bool VEffectiveAreaCalculator::getMonteCarloSpectra( VEffectiveAreaCalculatorMCHistograms* iMC_histo )
{
    // get solid angle normalization
    // (important for calculation of effective ares vs wobble offsets using diffuse gamma-ray simulations)
    double iSolAngleNorm = getMCSolidAngleNormalization();
    cout << "VEffectiveAreaCalculator::getMonteCarloSpectra: solid angle normalization factor: " << iSolAngleNorm << endl;
    
    char hname[800];
    // loop over az bins
    // (MC az [-180., 180.])
    for( unsigned int i_az = 0; i_az < fVMinAz.size(); i_az++ )
    {
        // loop over all spectral index
        for( unsigned int s = 0; s < fVSpectralIndex.size(); s++ )
        {
            if( s < hVEmc.size() && i_az < hVEmc[s].size() )
            {
                sprintf( hname, "hVEmc_%d_%d", s, i_az );
                if( iMC_histo->getHistogram_Emc( i_az, s ) )
                {
                    hVEmc[s][i_az] = ( TH1D* )iMC_histo->getHistogram_Emc( i_az, s )->Clone( hname );
                    if( hVEmc[s][i_az] )
                    {
                        hVEmc[s][i_az]->Scale( iSolAngleNorm );
                    }
                    if( hVEmc[s][i_az] && fRunPara && fRunPara->fIgnoreFractionOfEvents > 0. )
                    {
                        hVEmc[s][i_az]->Scale( ( 1.0 - fRunPara->fIgnoreFractionOfEvents ) );
                    }
                }
                else
                {
                    hVEmc[s][i_az] = 0;
                }
            }
            if( s < hVEmcSWeight.size() && i_az < hVEmcSWeight[s].size() )
            {
                sprintf( hname, "hVEmcSWeight_%d_%d", s, i_az );
                if( iMC_histo->getHistogram_EmcWeight( i_az, s ) )
                {
                    hVEmcSWeight[s][i_az] = ( TProfile* )iMC_histo->getHistogram_EmcWeight( i_az, s )->Clone( hname );
                    if( hVEmcSWeight[s][i_az] )
                    {
                        hVEmcSWeight[s][i_az]->Scale( iSolAngleNorm );
                    }
                    if( hVEmcSWeight[s][i_az] && fRunPara && fRunPara->fIgnoreFractionOfEvents > 0. )
                    {
                        hVEmcSWeight[s][i_az]->Scale( ( 1.0 - fRunPara->fIgnoreFractionOfEvents ) );
                    }
                }
                else
                {
                    hVEmcSWeight[s][i_az] = 0;
                }
            }
        }
    }
    
    return true;
}

/*
 *
 *  CALLED FOR CALCULATION OF EFFECTIVE AREAS
 *
 */
bool VEffectiveAreaCalculator::fill( CData* d, VEffectiveAreaCalculatorMCHistograms* iMC_histo, unsigned int iMethod )
{
    bool bDebugCuts = false;          // lots of debug output
    
    // make sure that vectors are initialized
    unsigned int ize = 0;      // should always be zero
    if( ize >= fZe.size() )
    {
        cout << "VEffectiveAreaCalculator::fill error: vectors are not correctly initialized " << fZe.size() << "\t" << ize << endl;
        return false;
    }
    resetHistogramsVectors();
    
    // do not require successfull energy reconstruction
    if( fIgnoreEnergyReconstruction )
    {
        iMethod = 100;
    }
    
    //////////////////////////////////////////////////////////////////
    // total Monte Carlo core scatter area (depends on CORSIKA shower core scatter mode)
    fMC_ScatterArea = 0.;
    if( fScatterMode[ize] == "VIEWCONE" )
    {
        fMC_ScatterArea = fAreaRadius[ize] * fAreaRadius[ize] * TMath::Pi();
    }
    else if( fScatterMode[ize] == "FLAT" )
    {
        fMC_ScatterArea = fAreaRadius[ize] * fAreaRadius[ize] * TMath::Pi() * cos( fZe[ize] * TMath::DegToRad() );
    }
    else
    {
        cout << "VEffectiveAreaCalculator::fill ERROR: unknown CORSIKA scatter mode: " << fScatterMode[ize] << endl;
        return false;
    }
    // reset unique event counter
    //	fUniqueEventCounter.clear();
    int iSuccessfullEventStatistics = 0;
    
    //////////////////////////////////////////////////////////////////
    // print some run information
    cout << endl;
    cout << "calculating effective areas: " << endl;
    cout << "\t zenith angle: " << fZe[ize];
    cout << ", wobble offset (x,y): " << fXWobble[ize] << ", " << fYWobble[ize];
    cout << " (" << sqrt( fXWobble[ize]*fXWobble[ize] + fYWobble[ize]*fYWobble[ize] ) << " deg)" << endl;
    cout << "\t noise level: " << fNoise[ize] << " (pedvar: " << fPedVar[ize] << ")" << endl;
    cout << "\t area (" << fScatterMode[ize] << ") [m^2]: " << fMC_ScatterArea;
    cout << " (scatter radius " << fAreaRadius[ize] << " [m])" << endl;
    cout << "\t energy reconstruction method: " << iMethod << endl;
    if( fIsotropicArrivalDirections )
    {
        cout << "\t assuming isotropic arrival directions" << endl;
    }
    if( fRunPara && fRunPara->fIgnoreFractionOfEvents > 0. )
    {
        cout << "\t ignore first " << fRunPara->fIgnoreFractionOfEvents * 100. << " % of events" << endl;
    }
    
    cout << endl;
    if( fSpectralWeight )
    {
        fSpectralWeight->print();
    }
    else
    {
        cout << "(no specral weight given)" << endl;
    }
    cout << endl;
    
    // make sure that all data pointers exist
    if( !d || !iMC_histo )
    {
        cout << "VEffectiveAreaCalculator::fill error: no data tree or MC histograms: " << endl;
        cout << "Data tree: " << d << ", MC histograms: " << iMC_histo << endl;
        return false;
    }
    
    // spectral weight
    double i_weight = 1.;
    // reconstructed energy (TeV, log10)
    double eRec = 0.;
    double eRecLin = 0.;
    // MC energy (TeV, log10)
    double eMC = 0.;
    
    ////////////////////////////////////////////////////////////////////////////
    // get MC histograms
    if( !getMonteCarloSpectra( iMC_histo ) )
    {
        cout << "VEffectiveAreaCalculator::fill error while getting MC spectra" << endl;
        return false;
    }
    
    ////////////////////////////////////////////////////////////////////////////
    // reset cut statistics
    fCuts->resetCutStatistics();
    
    ///////////////////////////////////////////////////////
    // get full data set and loop over all entries
    ///////////////////////////////////////////////////////
    Long64_t d_nentries = d->fChain->GetEntries();
    Long64_t i_start = 0;
    if( fRunPara && fRunPara->fIgnoreFractionOfEvents > 0. )
    {
        i_start = ( Long64_t )( fRunPara->fIgnoreFractionOfEvents * d_nentries );
    }
    cout << "\t total number of data events: " << d_nentries << " (start at event " << i_start << ")" << endl;
    
    //--- for the CR normalisation filling Acceptance tree total number of simulated is needed
    //-- WARNING if the rule for the azimuth bin changes in VInstrumentResponseFunctionRunParameter the following line must be adapted!!!!
    unsigned int number_of_az_bin = fRunPara->fAzMin.size();
    // if no azimuth bin, all events are in bin 0. if azimuth bin, all event are in the last bin
    int az_bin_index = 0;
    if( number_of_az_bin > 0 )
    {
        az_bin_index = ( int ) number_of_az_bin - 1;
    }
    // loop over all events
    for( Long64_t i = i_start; i < d_nentries; i++ )
    {
        d->GetEntry( i );
        
        // update cut statistics
        fCuts->newEvent();
        
        if( bDebugCuts )
        {
            cout << "============================== " << endl;
            cout << "EVENT entry number " << i << endl;
        }
        
        // apply MC cuts
        if( bDebugCuts )
        {
            cout << "#0 CUT MC " << fCuts->applyMCXYoffCut( d->MCxoff, d->MCyoff, false ) << endl;
        }
        
        if( !fCuts->applyMCXYoffCut( d->MCxoff, d->MCyoff, true ) )
        {
            continue;
        }
        
        // log of MC energy
        eMC = log10( d->MCe0 );
        
        // fill trigger cuts
        fillEcutSub( eMC, 0 );
        
        ////////////////////////////////
        // apply general quality and gamma/hadron separation cuts
        
        // apply reconstruction cuts
        if( bDebugCuts )
        {
            cout << "#1 CUT applyInsideFiducialAreaCut ";
            cout << fCuts->applyInsideFiducialAreaCut();
            cout << "\t" << fCuts->applyStereoQualityCuts( iMethod, false, i, true ) << endl;
        }
        
        // apply fiducial area cuts
        if( !fCuts->applyInsideFiducialAreaCut( true ) )
        {
            continue;
        }
        fillEcutSub( eMC, 1 );
        
        // apply reconstruction quality cuts
        if( !fCuts->applyStereoQualityCuts( iMethod, true, i , true ) )
        {
            continue;
        }
        fillEcutSub( eMC, 2 );
        
        // apply telescope type cut (e.g. for CTA simulations)
        if( fTelescopeTypeCutsSet )
        {
            if( bDebugCuts )
            {
                cout << "#2 Cut NTELType " << fCuts->applyTelTypeTest( false ) << endl;
            }
            if( !fCuts->applyTelTypeTest( true ) )
            {
                continue;
            }
        }
        fillEcutSub( eMC, 3 );
        
        
        //////////////////////////////////////
        // apply direction cut
        //
        // point source cut; use MC shower direction as reference direction
        bool bDirectionCut = false;
        if( !fIsotropicArrivalDirections )
        {
            if( !fCuts->applyDirectionCuts( iMethod, true ) )
            {
                bDirectionCut = true;
            }
        }
        // background cut; use (0,0) as reference direction
        // (command line option -d)
        else
        {
            if( !fCuts->applyDirectionCuts( iMethod, true, 0., 0. ) )
            {
                bDirectionCut = true;
            }
        }
        if( !bDirectionCut )
        {
            fillEcutSub( eMC, 4 );
        }
        
        //////////////////////////////////////
        // apply energy reconstruction quality cut
        if( !fIgnoreEnergyReconstruction )
        {
            if( bDebugCuts )
            {
                cout << "#4 EnergyReconstructionQualityCuts " << fCuts->applyEnergyReconstructionQualityCuts( iMethod ) << endl;
            }
            if( !fCuts->applyEnergyReconstructionQualityCuts( iMethod, true ) )
            {
                continue;
            }
        }
        if( !bDirectionCut )
        {
            fillEcutSub( eMC, 5 );
        }
        
        // skip event if no energy has been reconstructed
        // get energy according to reconstruction method
        if( fIgnoreEnergyReconstruction )
        {
            eRec = log10( d->MCe0 );
            eRecLin = d->MCe0;
        }
        else if( d->getEnergy_TeV() > 0. )
        {
            eRec = d->getEnergy_Log10();
            eRecLin = d->getEnergy_TeV();
        }
        else
        {
            continue;
        }
        
        /////////////////////////////////////////////////////////
        // fill response matrix after quality and direction cuts
        
        if( !bDirectionCut )
        {
            // loop over all az bins
            for( unsigned int i_az = 0; i_az < fVMinAz.size(); i_az++ )
            {
                if( !testAzimuthInterval( d, fZe[ize], fVMinAz[i_az], fVMaxAz[i_az] ) )
                {
                    continue;
                }
                // loop over all spectral index
                for( unsigned int s = 0; s < fVSpectralIndex.size(); s++ )
                {
                    if( hVResponseMatrixQC[s][i_az] )
                    {
                        hVResponseMatrixQC[s][i_az]->Fill( eRec, eMC );
                    }
                    if( hVResponseMatrixFineQC[s][i_az] )
                    {
                        hVResponseMatrixFineQC[s][i_az]->Fill( eRec, eMC );
                    }
                }
            }
        }
        
        //////////////////////////////////////
        // apply gamma hadron cuts
        if( bDebugCuts )
        {
            cout << "#3 CUT ISGAMMA " << fCuts->isGamma( i ) << endl;
        }
        if( !fCuts->isGamma( i, true ) )
        {
            continue;
        }
        if( !bDirectionCut )
        {
            fillEcutSub( eMC, 6 );
        }
        
        // unique event counter
        // (make sure that map doesn't get too big)
        if( !bDirectionCut && iSuccessfullEventStatistics >= 0 )
        {
            iSuccessfullEventStatistics++;
        }
        
        // loop over all az bins
        for( unsigned int i_az = 0; i_az < fVMinAz.size(); i_az++ )
        {
            if( !testAzimuthInterval( d, fZe[ize], fVMinAz[i_az], fVMaxAz[i_az] ) )
            {
                continue;
            }
            
            //fill tree with acceptance information after cuts (needed to construct background model in ctools)
            //NOTE: This tree is currently allways filled with the eventdisplay reconstruction results. If you want to have frogs/3Dmodel results, you need to fix that.
            if( !bDirectionCut && fRunPara->fgetXoff_Yoff_afterCut )
            {
                fXoff_aC = d->Xoff;
                fYoff_aC = d->Yoff;
                fXoff_derot_aC = d->Xoff_derot;
                fYoff_derot_aC = d->Yoff_derot;
                fErec = eRecLin;
                fEMC  = d->MCe0;
                fCRweight = getCRWeight( d->MCe0, hVEmc[0][az_bin_index], true ); //So that the acceptance can be normalised to the CR spectrum.
                // when running on gamma, this should return 1.
                fAcceptance_AfterCuts_tree->Fill();
            }
            
            
            // loop over all spectral index
            for( unsigned int s = 0; s < fVSpectralIndex.size(); s++ )
            {
                // weight by spectral index
                if( fSpectralWeight )
                {
                    fSpectralWeight->setSpectralIndex( fVSpectralIndex[s] );
                    i_weight = fSpectralWeight->getSpectralWeight( d->MCe0 );
                }
                else
                {
                    i_weight = 0.;
                }
                
                ////////////////////////////////////////////
                // fill effective areas before direction cut
                if( hVEcutNoTh2[s][i_az] )
                {
                    hVEcutNoTh2[s][i_az]->Fill( eMC, i_weight );
                }
                if( hVEcutRecNoTh2[s][i_az] )
                {
                    hVEcutRecNoTh2[s][i_az]->Fill( eRec, i_weight );
                }
                
                /////////////////////////
                // apply direction cut
                if( bDirectionCut )
                {
                    continue;
                }
                
                ///////////////////////////////////////////////////////////
                // from here on: after gamma/hadron and after direction cut
                
                // fill true MC energy (hVEmc is in true MC energies)
                if( hVEcut[s][i_az] )
                {
                    hVEcut[s][i_az]->Fill( eMC, i_weight );
                }
                if( hVEcutUW[s][i_az] )
                {
                    hVEcutUW[s][i_az]->Fill( eMC, 1. );
                }
                if( hVEcut500[s][i_az] )
                {
                    hVEcut500[s][i_az]->Fill( eMC, i_weight );
                }
                if( hVEcutRec[s][i_az] )
                {
                    hVEcutRec[s][i_az]->Fill( eRec, i_weight );
                }
                if( hVEcutRecUW[s][i_az] )
                {
                    hVEcutRecUW[s][i_az]->Fill( eRec, 1. );
                }
                if( hVEsysRec[s][i_az] )
                {
                    hVEsysRec[s][i_az]->Fill( eRec, eRec - eMC );
                }
                if( hVEsysMC[s][i_az] )
                {
                    hVEsysMC[s][i_az]->Fill( eMC, eRec - eMC );
                }
                if( hVEsysMCRelative[s][i_az] )
                {
                    hVEsysMCRelative[s][i_az]->Fill( eMC, ( eRecLin - d->MCe0 ) / d->MCe0 );
                }
                if( hVEsysMCRelativeRMS[s][i_az] )
                {
                    hVEsysMCRelativeRMS[s][i_az]->Fill( eMC, ( eRecLin - d->MCe0 ) / d->MCe0 );
                }
                if( hVEsysMCRelative2D[s][i_az] )
                {
                    hVEsysMCRelative2D[s][i_az]->Fill( eMC, eRecLin / d->MCe0 );
                }
                if( hVEsys2D[s][i_az] )
                {
                    hVEsys2D[s][i_az]->Fill( eMC, eRec - eMC );
                }
                if( hVResponseMatrix[s][i_az] )
                {
                    hVResponseMatrix[s][i_az]->Fill( eRec, eMC );
                }
                if( hVResponseMatrixFine[s][i_az] )
                {
                    hVResponseMatrixFine[s][i_az]->Fill( eRec, eMC, i_weight );
                }
                // events weighted by CR spectra
                if( hVWeightedRate[s][i_az] )
                {
                    hVWeightedRate[s][i_az]->Fill( eRec, getCRWeight( d->MCe0, hVEmc[s][i_az],
                                                   false, hVWeightedRate[s][i_az] ) );
                }
            }
        }
        // don't do anything between here and the end of the loop! Never!
    }  // end of loop
    /////////////////////////////////////////////////////////////////////////////
    
    
    
    /////////////////////////////////////////////////////////////////////////////
    //
    // calculate effective areas and fill output trees
    //
    /////////////////////////////////////////////////////////////////////////////
    
    ze = fZe[ize];
    fTNoise = fNoise[ize];
    // WARNING: hardwired values - not used to my knowledge anywhere? (GM)
    fTNoisePE = ( double )( fNoise[ize] ) / 0.15 * 1.e9;
    fTPedvar = fPedVar[ize];
    fXoff = fXWobble[ize];
    fYoff = fYWobble[ize];
    fWoff = sqrt( fXoff * fXoff + fYoff * fYoff );
    
    // loop over all spectral index
    for( unsigned int s = 0; s < fVSpectralIndex.size(); s++ )
    {
        fSpectralIndex = fVSpectralIndex[s];
        // loop over all az bins
        for( unsigned int i_az = 0; i_az < fVMinAz.size(); i_az++ )
        {
            fAzBin = ( int )i_az;
            fMinAz = fVMinAz[i_az];
            fMaxAz = fVMaxAz[i_az];
            
            // bayesdivide works only for weights == 1
            // errors might be wrong, since histograms are filled with weights != 1
            if( !binomialDivide( gEffAreaMC, hVEcut[s][i_az], hVEmc[s][i_az] ) )
            {
                cout << "VEffectiveAreaCalculator::fill: error calculating effective area vs MC energy" << endl;
                cout << "s : " << s << " , az: " << i_az << endl;
            }
            if( !binomialDivide( gEffAreaRec, hVEcutRec[s][i_az], hVEmc[s][i_az] ) )
            {
                cout << "VEffectiveAreaCalculator::fill: error calculating effective area vs rec energy" << endl;
                cout << "s : " << s << " , az: " << i_az << endl;
            }
            if( !binomialDivide( gEffAreaNoTh2MC, hVEcutNoTh2[s][i_az], hVEmc[s][i_az] ) )
            {
                cout << "VEffectiveAreaCalculator::fill: error calculating effective area vs MC energy" << endl;
                cout << "s : " << s << " , az: " << i_az << endl;
            }
            if( !binomialDivide( gEffAreaNoTh2Rec, hVEcutRecNoTh2[s][i_az], hVEmc[s][i_az] ) )
            {
                cout << "VEffectiveAreaCalculator::fill: error calculating effective area vs rec energy" << endl;
                cout << "s : " << s << " , az: " << i_az << endl;
            }
            // normalize all response matrices
            VHistogramUtilities::normalizeTH2D_y( hVResponseMatrix[s][i_az] );
            VHistogramUtilities::normalizeTH2D_y( hVResponseMatrixQC[s][i_az] );
            
            for( int i = 0; i < 1000; i++ )
            {
                e0[i] = 0.;
                eff[i] = 0.;
                eff_error[i] = 0.;
                esys_rel[i] = 0.;
                seff_L[i] = 0.;
                seff_U[i] = 0.;
                Rec_e0[i] = 0.;
                Rec_eff[i] = 0.;
                Rec_eff_error[i] = 0.;
                Rec_seff_L[i] = 0.;
                Rec_seff_U[i] = 0.;
                Rec_angRes_p68[i] = 0.;
                Rec_angRes_p80[i] = 0.;
            }
            double x = 0.;
            double y = 0.;
            // effective area vs MC energy
            nbins = gEffAreaMC->GetN();
            for( int i = 0; i < nbins; i++ )
            {
                gEffAreaMC->GetPoint( i, x, y );
                e0[i] = x;
                eff[i] = y * fMC_ScatterArea;
                seff_L[i] = gEffAreaMC->GetErrorYlow( i ) * fMC_ScatterArea;
                seff_U[i] = gEffAreaMC->GetErrorYhigh( i ) * fMC_ScatterArea;
                eff_error[i] = 0.5 * ( seff_L[i] + seff_U[i] );
                gEffAreaMC->SetPoint( i, x, eff[i] );
                gEffAreaMC->SetPointEYlow( i, seff_L[i] );
                gEffAreaMC->SetPointEYhigh( i, seff_U[i] );
                if( hVEsysMCRelative[s][i_az] )
                {
                    esys_rel[i] = hVEsysMCRelative[s][i_az]->GetBinContent( hVEsysMCRelative[s][i_az]->GetXaxis()->FindBin( e0[i] ) );
                }
            }
            
            // effective area vs reconstructed energy
            Rec_nbins = gEffAreaRec->GetN();
            for( int i = 0; i < Rec_nbins; i++ )
            {
                gEffAreaRec->GetPoint( i, x, y );
                Rec_e0[i] = x;
                // this is an approximation, since scatter area is defined over E_MC (GM: don't understand this comment)
                Rec_eff[i] = y * fMC_ScatterArea;
                Rec_seff_L[i] = gEffAreaRec->GetErrorYlow( i ) * fMC_ScatterArea;
                Rec_seff_U[i] = gEffAreaRec->GetErrorYhigh( i ) * fMC_ScatterArea;
                Rec_eff_error[i] = 0.5 * ( Rec_seff_L[i] + Rec_seff_U[i] );
                gEffAreaRec->SetPoint( i, x, Rec_eff[i] );
                gEffAreaRec->SetPointEYlow( i, Rec_seff_L[i] );
                gEffAreaRec->SetPointEYhigh( i, Rec_seff_U[i] );
            }
            multiplyByScatterArea( gEffAreaNoTh2MC );
            multiplyByScatterArea( gEffAreaNoTh2Rec );
            
            // copy all histograms
            resetHistograms( ize );
            copyHistograms( hEmc, hVEmc[s][i_az], false );
            copyHistograms( hEcut, hVEcut[s][i_az], false );
            copyHistograms( hEcutUW, hVEcutUW[s][i_az], false );
            copyHistograms( hEcut500, hVEcut500[s][i_az], false );
            copyHistograms( hEcutRec, hVEcutRec[s][i_az], false );
            copyHistograms( hEcutRecUW, hVEcutRecUW[s][i_az], false );
            copyProfileHistograms( hEmcSWeight, hVEmcSWeight[s][i_az] );
            copyProfileHistograms( hEsysRec,  hVEsysRec[s][i_az] );
            copyProfileHistograms( hEsysMC, hVEsysMC[s][i_az] );
            copyProfileHistograms( hEsysMCRelative, hVEsysMCRelative[s][i_az] );
            copyHistograms( hEsysMCRelativeRMS, hVEsysMCRelativeRMS[s][i_az], true );
            copyHistograms( hEsysMCRelative2D, hVEsysMCRelative2D[s][i_az], true );
            copyHistograms( hEsys2D, hVEsys2D[s][i_az], true );
            copyHistograms( hResponseMatrix, hVResponseMatrix[s][i_az], true );
            copyHistograms( hResponseMatrixQC, hVResponseMatrixQC[s][i_az], true );
            copyHistograms( hResponseMatrixFine, hVResponseMatrixFine[s][i_az], true );
            copyHistograms( hResponseMatrixFineQC, hVResponseMatrixFineQC[s][i_az], true );
            copyHistograms( hWeightedRate, hVWeightedRate[s][i_az], false );
            for( unsigned int e = 0; e < hEcutSub.size(); e++ )
            {
                copyHistograms( hEcutSub[e], hVEcutSub[s][e][i_az], false );
            }
            
            // fill angular resolution vs energy
            fillAngularResolution( i_az, false );
            fillAngularResolution( i_az, true );
            
            fEffArea->Fill();
        }
    }
    
    fCuts->printCutStatistics();
    
    // print out uniqueness of events
    /*    cout << "event statistics: " << endl;
        if( iSuccessfullEventStatistics > 0 )
        {
           map< unsigned int, unsigned short int>::iterator it;
           for( it = fUniqueEventCounter.begin(); it != fUniqueEventCounter.end(); it++ )
           {
    	  if( (*it).second > 1 )
    	  {
    	     cout << "\t multiple event after cuts at " << (double)((*it).first)/1.e5 << " TeV, used " << (*it).second << " times" << endl;
    	  }
           }
        }
        else iSuccessfullEventStatistics *= -1; */
    if( iSuccessfullEventStatistics < 0 )
    {
        iSuccessfullEventStatistics *= -1;
    }
    cout << "\t total number of events after cuts: " << iSuccessfullEventStatistics << endl;
    
    return true;
}


/*!
 *
 *  CALLED TO USE EFFECTIVE AREAS
 *
 * interpolate between effective area from different zenith angles with cos weights
 *
 *   return value is 1/effective area
 *
 *
 */
double VEffectiveAreaCalculator::getEffectiveArea( double erec, double ze, double woff, double iPedVar, double iSpectralIndex,
        bool bAddtoMeanEffectiveArea )
{
    if( bNOFILE )
    {
        return 1.;
    }
    
    // read effective areas from histograms
    return getEffectiveAreasFromHistograms( erec, ze, woff, iPedVar, iSpectralIndex, bAddtoMeanEffectiveArea );
}



/*
      this function always returns a vector of size 2
*/
vector< unsigned int > VEffectiveAreaCalculator::getUpperLowBins( vector< double > i_values, double d )
{
    vector< unsigned int > i_temp( 2, 0 );
    
    double i_min = 1.e10;
    unsigned int i_min_index = 0;
    double i_max = -1.e10;
    unsigned int i_max_index = 0;
    
    for( unsigned int i = 0; i < i_values.size(); i++ )
    {
        if( i_values[i] < i_min )
        {
            i_min = i_values[i];
            i_min_index = i;
        }
        if( i_values[i] > i_max )
        {
            i_max = i_values[i];
            i_max_index = i;
        }
    }
    if( d < i_min )
    {
        i_temp[0] = i_min_index;
        i_temp[1] = i_min_index;
        return i_temp;
    }
    if( d > i_max )
    {
        i_temp[0] = i_max_index;
        i_temp[1] = i_max_index;
        return i_temp;
    }
    
    // look for closest values
    double i_diff_upper = 1.e10;
    unsigned int i_diff_upper_index = 0;
    double i_diff_lower = 1.e10;
    unsigned int i_diff_lower_index = 0;
    for( unsigned int i = 0; i < i_values.size(); i++ )
    {
        if( i_values[i] - d > 0. && i_values[i] - d < i_diff_upper )
        {
            i_diff_upper = i_values[i] - d;
            i_diff_upper_index = i;
        }
        if( d - i_values[i] > 0. && d - i_values[i] < i_diff_lower )
        {
            i_diff_lower = d - i_values[i];
            i_diff_lower_index = i;
        }
    }
    i_temp[0] = i_diff_lower_index;
    i_temp[1] = i_diff_upper_index;
    
    return i_temp;
}


/*!
 *  CALLED TO USE EFFECTIVE AREAS
 *
 *  return effective area value for given ze, woff, iPedVar, ...
 *
 *
 */
double VEffectiveAreaCalculator::getEffectiveAreasFromHistograms( double erec, double ze, double woff,
        double iPedVar, double iSpectralIndex,
        bool bAddtoMeanEffectiveArea )
{
    vector< float > i_eff_temp( fEff_E0.size(), 0. );
    
    // log10 of energy
    if( erec <= 0. )
    {
        return 0.;
    }
    float lerec = log10( erec );
    
    // calculate mean values
    fEffectiveAreas_meanZe     += ze;
    fEffectiveAreas_meanWoff   += woff;
    fEffectiveAreas_meanPedVar += iPedVar;
    fEffectiveAreas_meanIndex   = iSpectralIndex;
    fEffectiveAreas_meanN++;
    
    ////////////////////////////////////////////////////////
    // get upper and lower zenith angle bins
    ////////////////////////////////////////////////////////
    vector< unsigned int > i_ze_bins = getUpperLowBins( fZe, ze );
    vector< vector< float > > i_ze_eff_temp( 2, i_eff_temp );
    for( unsigned int i = 0; i < i_ze_bins.size(); i++ )
    {
        ////////////////////////////////////////////////////////
        // get lower and upper wobble offset bins
        ////////////////////////////////////////////////////////
        if( i_ze_bins[i] < fEff_WobbleOffsets.size() )
        {
            vector< unsigned int > i_woff_bins = getUpperLowBins( fEff_WobbleOffsets[i_ze_bins[i]], woff );
            vector< vector< float > > i_woff_eff_temp( 2, i_eff_temp );
            for( unsigned int w = 0; w < i_woff_bins.size(); w++ )
            {
                ////////////////////////////////////////////////////////
                // get lower and upper noise bins
                ////////////////////////////////////////////////////////
                if( i_ze_bins[i] < fEff_Noise.size() && i_woff_bins[w] < fEff_Noise[i_ze_bins[i]].size() )
                {
                    vector< unsigned int > i_noise_bins = getUpperLowBins( fEff_Noise[i_ze_bins[i]][i_woff_bins[w]], iPedVar );
                    vector< vector< float > > i_noise_eff_temp( 2, i_eff_temp );
                    for( unsigned int n = 0; n < i_noise_bins.size(); n++ )
                    {
                        ////////////////////////////////////////////////////////
                        // get lower and upper spectral index bins
                        ////////////////////////////////////////////////////////
                        if( i_ze_bins[i] < fEff_SpectralIndex.size()
                                && i_woff_bins[w] < fEff_SpectralIndex[i_ze_bins[i]].size()
                                && i_noise_bins[n] < fEff_SpectralIndex[i_ze_bins[i]][i_woff_bins[w]].size() )
                        {
                            vector< unsigned int > i_index_bins = getUpperLowBins( fEff_SpectralIndex[i_ze_bins[i]][i_woff_bins[w]][i_noise_bins[n]],
                                                                  iSpectralIndex );
                            unsigned int i_ID_0 = i_index_bins[0] + 100 * ( i_noise_bins[n] + 100 * ( i_woff_bins[w] + 100 * i_ze_bins[i] ) );
                            unsigned int i_ID_1 = i_index_bins[1] + 100 * ( i_noise_bins[n] + 100 * ( i_woff_bins[w] + 100 * i_ze_bins[i] ) );
                            i_noise_eff_temp[n] = interpolate_effectiveArea( iSpectralIndex,
                                                  fEff_SpectralIndex[i_ze_bins[i]][i_woff_bins[w]][i_noise_bins[n]][i_index_bins[0]],
                                                  fEff_SpectralIndex[i_ze_bins[i]][i_woff_bins[w]][i_noise_bins[n]][i_index_bins[1]],
                                                  fEffArea_map[i_ID_0],
                                                  fEffArea_map[i_ID_1], false );
                        }
                        else
                        {
                            cout << "VEffectiveAreaCalculator::getEffectiveAreasFromHistograms error: spectral index index out of range: ";
                            cout << i_ze_bins[i] << " " << fEff_SpectralIndex.size();
                            if( i_ze_bins[i] < fEff_SpectralIndex.size() )
                            {
                                cout << " " << i_woff_bins[w] << " " << fEff_SpectralIndex[i_ze_bins[i]].size();
                            }
                            if( i_woff_bins[w] < fEff_SpectralIndex[i_ze_bins[i]].size() )
                            {
                                cout << " " << i_noise_bins[n] <<  fEff_SpectralIndex[i_ze_bins[i]][i_woff_bins[w]].size();
                            }
                            cout << endl;
                            return -1.;
                        }
                        ////////////////////////////////////////////////////////
                    }
                    i_woff_eff_temp[w] = interpolate_effectiveArea( iPedVar,
                                         fEff_Noise[i_ze_bins[i]][i_woff_bins[w]][i_noise_bins[0]],
                                         fEff_Noise[i_ze_bins[i]][i_woff_bins[w]][i_noise_bins[1]],
                                         i_noise_eff_temp[0],
                                         i_noise_eff_temp[1], false );
                }
                else
                {
                    cout << "VEffectiveAreaCalculator::getEffectiveAreasFromHistograms error: noise index out of range: ";
                    cout << i_ze_bins[i] << " " << fEff_Noise.size();
                    if( i_ze_bins[i] < fEff_Noise.size() )
                    {
                        cout << " " << i_woff_bins[w] << " " << fEff_Noise[i_ze_bins[i]].size() << endl;
                    }
                    cout << endl;
                    return -1.;
                }
            }
            i_ze_eff_temp[i] = interpolate_effectiveArea( woff,
                               fEff_WobbleOffsets[i_ze_bins[i]][i_woff_bins[0]],
                               fEff_WobbleOffsets[i_ze_bins[i]][i_woff_bins[1]],
                               i_woff_eff_temp[0],
                               i_woff_eff_temp[1], false );
        }
        else
        {
            cout << "VEffectiveAreaCalculator::getEffectiveAreasFromHistograms error: woff index out of range: ";
            cout << i_ze_bins[i] << " " << fEff_WobbleOffsets.size() << endl;
            return -1.;
        }
    }
    i_eff_temp = interpolate_effectiveArea( ze, fZe[i_ze_bins[0]], fZe[i_ze_bins[1]], i_ze_eff_temp[0], i_ze_eff_temp[1], true );
    
    if( fEff_E0.size() == 0 )
    {
        return -1.;
    }
    
    // mean effective area calculation
    if( bAddtoMeanEffectiveArea && fVTimeBinnedMeanEffectiveArea.size() == i_eff_temp.size() )
    {
        for( unsigned int i = 0; i < i_eff_temp.size(); i++ )
        {
            fVTimeBinnedMeanEffectiveArea[i] += i_eff_temp[i];
        }
        fNTimeBinnedMeanEffectiveArea++;
    }
    
    /////////////////////////////////////////
    // effective area for a specific energy
    /////////////////////////////////////////
    float i_eff_e = 1.;
    unsigned int ie0_low = 0;
    unsigned int ie0_up = 0;
    
    if( lerec <= fEff_E0[0] )
    {
        ie0_low = ie0_up = 0;
    }
    else if( lerec > fEff_E0[fEff_E0.size() - 1] )
    {
        ie0_low = ie0_up = fEff_E0.size() - 1;
    }
    else
    {
        for( unsigned int j = 1; j < fEff_E0.size() - 1; j++ )
        {
            if( lerec > fEff_E0[j] )
            {
                ie0_low = j;
                ie0_up = j + 1;
            }
        }
    }
    
    ///////////////////////////////////
    // linear interpolate between energies
    ///////////////////////////////////
    
    i_eff_e = VStatistics::interpolate( i_eff_temp[ie0_low], fEff_E0[ie0_low], i_eff_temp[ie0_up], fEff_E0[ie0_up], lerec, false );
    
    if( i_eff_e > 0. )
    {
        return 1. / i_eff_e;
    }
    
    return -1.;
}

// reset the sum of effective areas

void VEffectiveAreaCalculator::resetTimeBin()
{
    for( unsigned int i = 0; i < fVTimeBinnedMeanEffectiveArea.size(); i++ )
    {
        fVTimeBinnedMeanEffectiveArea[i] = 0;
    }
    fNTimeBinnedMeanEffectiveArea = 0;
}


/*
 *  CALLED FOR CALCULATION OF EFFECTIVE AREAS
 */
void VEffectiveAreaCalculator::resetHistograms( unsigned int ize )
{
    if( ize >= fZe.size() )
    {
        return;
    }
    
    char htitle[400];
    if( hisTreeListofHistograms )
    {
        TIter next( hisTreeListofHistograms );
        while( TH1* obj = ( TH1* )next() )
        {
            obj->Reset();
            string iTemp = obj->GetTitle();
            if( iTemp.find( "(" ) != string::npos )
            {
                iTemp = iTemp.substr( 0, iTemp.find( "(" ) - 1 );
            }
            sprintf( htitle, "%s (%1.f deg)", iTemp.c_str(), fZe[ize] );
            obj->SetTitle( htitle );
        }
    }
    
    // set titles of effective area graphs
    sprintf( htitle, "effective area vs E_{MC} (%.1f deg)", fZe[ize] );
    gEffAreaMC->SetTitle( htitle );
    
    sprintf( htitle, "effective area vs E_{rec} (%.1f deg)", fZe[ize] );
    gEffAreaRec->SetTitle( htitle );
    
    sprintf( htitle, "effective area vs E_{MC} (%.1f deg)", fZe[ize] );
    gEffAreaNoTh2MC->SetTitle( htitle );
    
    sprintf( htitle, "effective area vs E_{rec} (%.1f deg)", fZe[ize] );
    gEffAreaNoTh2Rec->SetTitle( htitle );
}


void VEffectiveAreaCalculator::setWobbleOffset( double x, double y )
{
    if( fXWobble.size() == 0 )
    {
        fXWobble.push_back( x );
    }
    else
    {
        fXWobble[0] = x;
    }
    if( fYWobble.size() == 0 )
    {
        fYWobble.push_back( y );
    }
    else
    {
        fYWobble[0] = y;
    }
    
}

void VEffectiveAreaCalculator::setNoiseLevel( int iN, double iP )
{
    if( fNoise.size() == 0 )
    {
        fNoise.push_back( iN );
    }
    else
    {
        fNoise[0] = iN;
    }
    
    if( fPedVar.size() == 0 )
    {
        fPedVar.push_back( iP );
    }
    else
    {
        fPedVar[0] = iP;
    }
}

/*!
    set azimuth cut for effective area fillings
*/
void VEffectiveAreaCalculator::setAzimuthCut( int iAzBin, double iAzMin, double iAzMax )
{
    fAzBin = iAzBin;
    fMinAz = iAzMin;
    fMaxAz = iAzMax;
}


bool VEffectiveAreaCalculator::binomialDivide( TGraphAsymmErrors* g, TH1D* hrec, TH1D* hmc )
{
    if( !g )
    {
        cout << "VEffectiveAreaCalculator::binomialDivide error: no return graph given" << endl;
        return false;
    }
    if( !hrec )
    {
        cout << "VEffectiveAreaCalculator::binomialDivide error: no histogram with reconstructed events given" << endl;
        return false;
    }
    if( !hmc )
    {
        cout << "VEffectiveAreaCalculator::binomialDivide error: no histogram with simulated events given" << endl;
        return false;
    }
    
    int z = 0;
    double pj = 0.;
    double pr = 0.;
    double pm = 0.;
    double sj_low = 0.;
    double sj_up = 0.;
    
    for( int b = 1; b <= hmc->GetNbinsX(); b++ )
    {
        if( hmc->GetBinContent( b ) > 0 && hrec->GetBinContent( b ) > 0 )
        {
            pj = hrec->GetBinContent( b ) / hmc->GetBinContent( b );
            // error calculation for effective areas
            //  this far from being straightforward!
            //  none of the methods works consistently, therefore the simplest (normal) solution
            //  is used.
            //  note: approach is only correct for unweighted histograms
            //
            // error calculation assuming binomial distributions
            // (see Blobel/Lohrmann; chapter 11.2 (Akzeptanzkorrekturen)
            pr = hrec->GetBinError( b );
            pm = hmc->GetBinError( b );
            if( pj != 1. )
            {
                sj_low = TMath::Abs( ( ( 1. - 2.*pj ) * pr * pr + pj * pj * pm * pm ) / ( hmc->GetBinContent( b ) * hmc->GetBinContent( b ) ) );
            }
            else
            {
                sj_low = 0.;
            }
            sj_low = sqrt( sj_low );
            sj_up  = sj_low;
            // fill effective area graphs
            g->SetPoint( z, hmc->GetBinCenter( b ), pj );
            g->SetPointError( z, 0., 0., sj_low, sj_up );
            z++;
        }
    }
    g->Set( z );
    
    return true;
}

/*

    running mean smoothing of effective area curve
    (not default)

    needs careful testing before using
*/
void VEffectiveAreaCalculator::smoothEffectiveAreas( map< unsigned int, vector< float > > m )
{
    cout << "\t smooth effective areas, parameters: " << fSmoothIter << ", " << fSmoothThreshold << endl;
    
    typedef map< unsigned int, vector< float > >::const_iterator CI;
    
    for( CI p = m.begin(); p != m.end(); ++ p )
    {
        vector< float > itemp = p->second;
        
        for( int l = 0; l < fSmoothIter; l++ )
        {
            for( unsigned int k = 1; k < itemp.size() - 1; k++ )
            {
                if( itemp[k - 1] <= 0 || itemp[k] <= 0. || itemp[k + 1] <= 0. )
                {
                    continue;
                }
                
                // upwards outlier
                if( itemp[k] / itemp[k - 1] > ( 1. + fSmoothThreshold ) && itemp[k] / itemp[k + 1] > ( 1. + fSmoothThreshold ) )
                {
                    itemp[k] = 0.5 * ( itemp[k - 1] + itemp[k + 1] );
                }
                
                // downwards outlier
                if( itemp[k] / itemp[k - 1] < ( 1. - fSmoothThreshold ) && itemp[k] / itemp[k + 1] < ( 1. - fSmoothThreshold ) )
                {
                    itemp[k] = 0.5 * ( itemp[k - 1] + itemp[k + 1] );
                }
            }
        }
    }
}


void VEffectiveAreaCalculator::copyProfileHistograms( TProfile* h1,  TProfile* h2 )
{
    if( !h1 || !h2 )
    {
        return;
    }
    
    string iEOption = h1->GetErrorOption();
    
    for( int b = 0; b <= h2->GetNbinsX(); b++ )
    {
        h1->SetBinContent( b, h2->GetBinContent( b ) * h2->GetBinEntries( b ) );
        
        if( TMath::Abs( h2->GetBinError( b ) ) < 1.e-4 )
        {
            h1->SetBinError( b, 0. );
        }
        else
        {
            if( h2->GetBinEntries( b ) > 0. )
            {
                double iE = h2->GetBinError( b );
                if( iEOption != "s" )
                {
                    iE = h2->GetBinError( b ) * sqrt( h2->GetBinEntries( b ) );
                }
                h1->SetBinError( b,  sqrt( h2->GetBinEntries( b ) * ( h2->GetBinContent( b ) *  h2->GetBinContent( b ) + iE * iE ) ) );
            }
            else
            {
                h1->SetBinError( b, 0. );
            }
        }
        h1->SetBinEntries( b, h2->GetBinEntries( b ) );
    }
    h1->SetEntries( h2->GetEntries() );
}


void VEffectiveAreaCalculator::copyHistograms( TH1* h1,  TH1* h2, bool b2D )
{
    if( !h1 || !h2 )
    {
        return;
    }
    
    if( !b2D )
    {
        for( int b = 0; b <= h2->GetNbinsX(); b++ )
        {
            h1->SetBinContent( b, h2->GetBinContent( b ) );
            h1->SetBinError( b, h2->GetBinError( b ) );
        }
    }
    else
    {
        for( int b = 0; b <= h2->GetNbinsX(); b++ )
        {
            for( int b2 = 0; b2 <= h2->GetNbinsY(); b2++ )
            {
                h1->SetBinContent( b, b2, h2->GetBinContent( b, b2 ) );
                h1->SetBinError( b, b2, h2->GetBinError( b, b2 ) );
            }
        }
    }
    h1->SetEntries( h2->GetEntries() );
}


/*

    return mean effective area for given run

    values are filled ....

*/
TGraphAsymmErrors* VEffectiveAreaCalculator::getMeanEffectiveArea()
{
    if( gMeanEffectiveArea && fEffArea_time.size() > 0 )
    {
        gMeanEffectiveArea->Set( 0 );
        double z = 0;
        double x = 0.;
        double y = 0.;
        
        // set energies
        // energy dimension of fEffArea_time is alway of same size
        for( unsigned int i = 0; i < fEff_E0.size(); i++ )
        {
            gMeanEffectiveArea->SetPoint( i, fEff_E0[i], 0. );
        }
        
        for( unsigned int i = 0; i < fEffArea_time.size(); i++ )
        {
            for( unsigned int j = 0; j < fEffArea_time[i].size(); j++ )
            {
                gMeanEffectiveArea->GetPoint( j, x, y );
                gMeanEffectiveArea->SetPoint( j, x, y + fEffArea_time[i][j] );
            }
        }
        
        z = ( double )fEffArea_time.size();
        for( int i = 0; i < gMeanEffectiveArea->GetN(); i++ )
        {
            if( z > 0. )
            {
                gMeanEffectiveArea->GetPoint( i, x, y );
                gMeanEffectiveArea->SetPoint( i, x, y / z );
            }
        }
        return gMeanEffectiveArea;
    }
    
    return 0;
}

/*

    return mean effective area for a given time bin


*/
TGraph2DErrors* VEffectiveAreaCalculator::getTimeBinnedMeanEffectiveArea()
{
    if( gTimeBinnedMeanEffectiveArea )
    {
        int z = 0;
        gTimeBinnedMeanEffectiveArea->Set( 0 );
        for( unsigned int i = 0; i < fEffArea_time.size(); i++ )
        {
            for( unsigned int j = 0; j < fEffArea_time[i].size(); j++ )
            {
                if( fEffArea_time[i][j] > 0. )
                {
                    gTimeBinnedMeanEffectiveArea->SetPoint( z , fEff_E0[j], fEffArea_time[i][j], timebins[i] );
                    gTimeBinnedMeanEffectiveArea->SetPointError( z , 0 , 0 , 0 );
                    z++;
                }
            }
        }
        return gTimeBinnedMeanEffectiveArea;
    }
    
    return 0;
}


/*

   set a new time bin and calculate mean effective area

*/
void VEffectiveAreaCalculator::setTimeBinnedMeanEffectiveArea( double i_time )
{
    timebins.push_back( i_time );
    
    vector < double > inter;
    
    for( unsigned int i = 0; i < fVTimeBinnedMeanEffectiveArea.size(); i++ )
    {
        if( fNTimeBinnedMeanEffectiveArea > 0. )
        {
            inter.push_back( fVTimeBinnedMeanEffectiveArea[i] / fNTimeBinnedMeanEffectiveArea );
        }
        else
        {
            inter.push_back( 0. );
        }
    }
    
    fEffArea_time.push_back( inter );
    
    resetTimeBin();
}


void VEffectiveAreaCalculator::resetHistogramsVectors( )
{
    if( hisVList )
    {
        return;
    }
    
    // loop over all temporary histograms and reset them
    TIter next( hisVList );
    while( TH1* obj = ( TH1* )next() )
    {
        obj->Reset();
    }
}


TH1D* VEffectiveAreaCalculator::getHistogramhEmc()
{
    if( !hEmc )
    {
        return 0;
    }
    
    hEmc->Reset();
    
    return hEmc;
}

/*
 * calculate mean systematic error (bias) in energy reconstruction
 *
 * interpolate betwen ze, az, noise, woff bins
 */
TGraphErrors* VEffectiveAreaCalculator::getMeanSystematicErrorHistogram()
{
    if( fEffectiveAreas_meanN <= 0. )
    {
        return 0;
    }
    fEffectiveAreas_meanZe /= fEffectiveAreas_meanN;
    fEffectiveAreas_meanWoff /= fEffectiveAreas_meanN;
    fEffectiveAreas_meanPedVar /= fEffectiveAreas_meanN;
    
    // new graph with mean systematic (bias) error in energy reconstruction
    gMeanSystematicErrorGraph = new TGraphErrors( 1 );
    gMeanSystematicErrorGraph->SetName( "gMeanEnergySystematicError" );
    gMeanSystematicErrorGraph->SetTitle( "" );
    gMeanSystematicErrorGraph->SetMarkerStyle( 20 );
    vector< float > hX;
    vector< float > i_eff_temp;
    
    ////////////////////////////////////////////////////////
    // get upper and lower zenith angle bins
    ////////////////////////////////////////////////////////
    vector< unsigned int > i_ze_bins = getUpperLowBins( fZe, fEffectiveAreas_meanZe );
    vector< vector< float > > i_ze_eff_temp( 2, hX );
    for( unsigned int i = 0; i < i_ze_bins.size(); i++ )
    {
        ////////////////////////////////////////////////////////
        // get lower and upper wobble offset bins
        ////////////////////////////////////////////////////////
        if( i_ze_bins[i] >= fEff_WobbleOffsets.size() )
        {
            continue;
        }
        vector< unsigned int > i_woff_bins = getUpperLowBins( fEff_WobbleOffsets[i_ze_bins[i]], fEffectiveAreas_meanWoff );
        vector< vector< float > > i_woff_eff_temp( 2, hX );
        for( unsigned int w = 0; w < i_woff_bins.size(); w++ )
        {
            ////////////////////////////////////////////////////////
            // get lower and upper noise bins
            ////////////////////////////////////////////////////////
            if( i_ze_bins[i] >= fEff_Noise.size() || i_woff_bins[w] >= fEff_Noise[i_ze_bins[i]].size() )
            {
                continue;
            }
            vector< unsigned int > i_noise_bins = getUpperLowBins( fEff_Noise[i_ze_bins[i]][i_woff_bins[w]], fEffectiveAreas_meanPedVar );
            vector< vector< float > > i_noise_eff_temp( 2, hX );
            for( unsigned int n = 0; n < i_noise_bins.size(); n++ )
            {
                ////////////////////////////////////////////////////////
                // get lower and upper spectral index bins
                ////////////////////////////////////////////////////////
                if( i_ze_bins[i] >= fEff_SpectralIndex.size() || i_woff_bins[w] >= fEff_SpectralIndex[i_ze_bins[i]].size()
                        || i_noise_bins[n] >= fEff_SpectralIndex[i_ze_bins[i]][i_woff_bins[w]].size() )
                {
                    continue;
                }
                vector< unsigned int > i_index_bins = getUpperLowBins( fEff_SpectralIndex[i_ze_bins[i]][i_woff_bins[w]][i_noise_bins[n]],
                                                      fEffectiveAreas_meanIndex );
                unsigned int i_ID_0 = i_index_bins[0] + 100 * ( i_noise_bins[n] + 100 * ( i_woff_bins[w] + 100 * i_ze_bins[i] ) );
                unsigned int i_ID_1 = i_index_bins[1] + 100 * ( i_noise_bins[n] + 100 * ( i_woff_bins[w] + 100 * i_ze_bins[i] ) );
                i_noise_eff_temp[n] = interpolate_effectiveArea(
                                          fEffectiveAreas_meanIndex,
                                          fEff_SpectralIndex[i_ze_bins[i]][i_woff_bins[w]][i_noise_bins[n]][i_index_bins[0]],
                                          fEff_SpectralIndex[i_ze_bins[i]][i_woff_bins[w]][i_noise_bins[n]][i_index_bins[1]],
                                          fEff_EsysMCRelative[i_ID_0], fEff_EsysMCRelative[i_ID_1], false );
                // don't interpolate errors, assume they are more or less constant
            }
            i_woff_eff_temp[w] = interpolate_effectiveArea( fEffectiveAreas_meanPedVar,
                                 fEff_Noise[i_ze_bins[i]][i_woff_bins[w]][i_noise_bins[0]],
                                 fEff_Noise[i_ze_bins[i]][i_woff_bins[w]][i_noise_bins[1]],
                                 i_noise_eff_temp[0], i_noise_eff_temp[1], false );
        }
        i_ze_eff_temp[i] = interpolate_effectiveArea( fEffectiveAreas_meanWoff,
                           fEff_WobbleOffsets[i_ze_bins[i]][i_woff_bins[0]],
                           fEff_WobbleOffsets[i_ze_bins[i]][i_woff_bins[1]],
                           i_woff_eff_temp[0], i_woff_eff_temp[1], false );
    }
    i_eff_temp = interpolate_effectiveArea( fEffectiveAreas_meanZe,
                                            fZe[i_ze_bins[0]],
                                            fZe[i_ze_bins[1]],
                                            i_ze_eff_temp[0], i_ze_eff_temp[1], true );
    if( fEff_E0.size() == i_eff_temp.size() )
    {
        unsigned int z = 0;
        for( unsigned int i = 0; i < i_eff_temp.size(); i++ )
        {
            // don't expect these values to be exactly zero
            if( TMath::Abs( i_eff_temp[i] ) > 1.e-8 )
            {
                gMeanSystematicErrorGraph->SetPoint( z, fEff_E0[i], i_eff_temp[i] );
                gMeanSystematicErrorGraph->SetPointError( z, 0., 0. );
                z++;
            }
        }
    }
    
    return gMeanSystematicErrorGraph;
}


bool VEffectiveAreaCalculator::setMonteCarloEnergyRange( double iMin, double iMax, double iMCIndex )
{
    if( fSpectralWeight )
    {
        fSpectralWeight->setMCParameter( iMCIndex, iMin, iMax );
        return true;
    }
    
    cout << "VEffectiveAreaCalculator::setMonteCarloEnergyRange error: not spectral weighter defined" << endl;
    return false;
}
/*
 * weight to correct the MC spectrum to the CR spectrum
 *
 * note that this requires the MC spectrum to be a power law, as it is integrated by hand
 *
 * weight derived from dN_CR (E) = w(E) * dN_MC (E)
 *
 */
double VEffectiveAreaCalculator::getCRWeight( double iEMC_TeV_lin, TH1* h, bool for_back_map, TH1* hF )
{
    if( !h || !fRunPara )
    {
        return 1.;
    }
    
    if( !fRunPara->fCREnergySpectrum )
    {
        return 1.;
    }
    
    // energy bin upper and lower bounds
    double iE_min = 0.;
    double iE_max = 0.;
    if( hF )
    {
        iE_min = hF->GetXaxis()->GetBinLowEdge( hF->GetXaxis()->FindBin( log10( iEMC_TeV_lin ) ) );
        iE_max = hF->GetXaxis()->GetBinUpEdge( hF->GetXaxis()->FindBin( log10( iEMC_TeV_lin ) ) );
        iE_min = TMath::Power( 10., iE_min );
        iE_max = TMath::Power( 10., iE_max );
    }
    
    // normalization of MC spectrum
    double c_ig = 0.;
    if( fRunPara->fIgnoreFractionOfEvents > 0. )
    {
        c_ig = fRunPara->fIgnoreFractionOfEvents;
    }
    double c_mc = ( 1.0 - c_ig ) * h->GetEntries()
                  * ( -1.*TMath::Abs( fRunPara->fMCEnergy_index ) + 1. )
                  / ( TMath::Power( fRunPara->fMCEnergy_max, -1.*TMath::Abs( fRunPara->fMCEnergy_index ) + 1. )
                      - TMath::Power( fRunPara->fMCEnergy_min, -1.*TMath::Abs( fRunPara->fMCEnergy_index ) + 1. ) );
                      
                      
    // number of MC events in this energy bin
    double n_mc = 1.;
    if( hF )
    {
        n_mc = c_mc / ( -1.*TMath::Abs( fRunPara->fMCEnergy_index ) + 1. )
               * ( TMath::Power( iE_max , -1.*TMath::Abs( fRunPara->fMCEnergy_index ) + 1. )
                   - TMath::Power( iE_min, -1.*TMath::Abs( fRunPara->fMCEnergy_index ) + 1. ) );
    }
    else
    {
        n_mc = c_mc * TMath::Power( iEMC_TeV_lin, -1.*TMath::Abs( fRunPara->fMCEnergy_index ) );
    }
    
    // number of expected CR events /min/sr (for gammas: /min)
    // (convert to cm2 and min)
    double n_cr = fMC_ScatterArea * 1.e4 * 60.;
    if( hF )
    {
        n_cr *= fRunPara->fCREnergySpectrum->Integral( iE_min, iE_max );
    }
    else
    {
        n_cr *= fRunPara->fCREnergySpectrum->Eval( log10( iEMC_TeV_lin ) );
    }
    
    
    // (ctools) for the acceptance map construction, the weight is in #/s ()
    if( for_back_map )
    {
        // need to normalize this number considering the cone in which the particle are simulated
        // (the offset bin is not consider so the binning can be changed later).
        if( !fsolid_angle_norm_done )
        {
            Calculate_Bck_solid_angle_norm();
        }
        n_cr = fsolid_angle_norm * n_cr / 60.;
        if( n_mc != 0. )
        {
            return n_cr / n_mc;
        }
        else
        {
            return 0.;
        }
    }
    else
    {
        // getMCSolidAngleNormalization() return a ratio of solid angle (only for gamma? not sure this thing returning something else than 1 here, ever...)
        if( getMCSolidAngleNormalization() > 0. )
        {
            n_cr /= getMCSolidAngleNormalization();    // do we want to leave this line here?
        }
        // the solid angle normalization is done in VSensitivityCalculator
        // return #/min/sr
        if( n_mc != 0. )
        {
            return n_cr / n_mc;
        }
        return 0.;
    }
}

// Calculate the ratio between the solid angle of the cone where the MC have been done and the solid angle of the offset bin considered
void VEffectiveAreaCalculator::Calculate_Bck_solid_angle_norm()
{


    if( fRunPara->fViewcone_max > 0. )
    {
        // solid angle in which the particule have been simulated
        double SolidAngle_MCScatterAngle  =  2 * TMath::Pi() * ( 1. - cos( fRunPara->fViewcone_max * TMath::DegToRad() ) );
        
        fsolid_angle_norm = SolidAngle_MCScatterAngle;
        fsolid_angle_norm_done = true;
        
    }
    
    
    return;
}

/*

    check if azimuth is in correct bin

*/
bool VEffectiveAreaCalculator::testAzimuthInterval( CData* d, double iZe, double iMinAz, double iMaxAz )
{
    if( !d )
    {
        return false;
    }
    
    // check at what azimuth bin we are
    if( iZe > 3. )
    {
        // confine MC az to -180., 180.
        if( d->MCaz > 180. )
        {
            d->MCaz -= 360.;
        }
        // expect bin like [135,-135]
        if( iMinAz > iMaxAz )
        {
            if( d->MCaz < iMinAz && d->MCaz > iMaxAz )
            {
                return false;
            }
        }
        // expect bin like [-135,-45.]
        else
        {
            if( d->MCaz < iMinAz || d->MCaz > iMaxAz )
            {
                return false;
            }
        }
    }
    return true;
}

void VEffectiveAreaCalculator::multiplyByScatterArea( TGraphAsymmErrors* g )
{
    if( !g )
    {
        return;
    }
    double x = 0.;
    double y = 0.;
    for( int i = 0; i < g->GetN(); i++ )
    {
        g->GetPoint( i, x, y );
        y *= fMC_ScatterArea;
        g->SetPoint( i, x, y );
        g->SetPointEYlow( i, g->GetErrorYlow( i ) * fMC_ScatterArea );
        g->SetPointEYhigh( i, g->GetErrorYhigh( i ) * fMC_ScatterArea );
    }
}

/*

   copy angular resolution values to tree variable

*/
void VEffectiveAreaCalculator::fillAngularResolution( unsigned int i_az, bool iContaintment_80p )
{
    if( iContaintment_80p && i_az < fGraph_AngularResolution68p.size() && fGraph_AngularResolution68p[i_az] )
    {
        // get first and last energy bin
        double i_emin = 1.e5;
        double i_emax = 1.e-5;
        double x = 0.;
        double y = 0.;
        for( int i = 0; i < fGraph_AngularResolution80p[i_az]->GetN(); i++ )
        {
            fGraph_AngularResolution80p[i_az]->GetPoint( i, x, y );
            if( x < i_emin )
            {
                i_emin = x;
            }
            if( x > i_emax )
            {
                i_emax = x;
            }
        }
        for( int i = 0; i < Rec_nbins; i++ )
        {
            if( Rec_e0[i] > i_emin && Rec_e0[i] < i_emax )
            {
                Rec_angRes_p80[i]  = fGraph_AngularResolution80p[i_az]->Eval( Rec_e0[i] );
            }
        }
    }
    else if( !iContaintment_80p && i_az < fGraph_AngularResolution68p.size() && fGraph_AngularResolution68p[i_az] )
    {
        double i_emin = 0.;
        double i_emax = 0.;
        double x = 0.;
        double y = 0.;
        for( int i = 0; i < fGraph_AngularResolution68p[i_az]->GetN(); i++ )
        {
            fGraph_AngularResolution68p[i_az]->GetPoint( i, x, y );
            if( x < i_emin )
            {
                i_emin = x;
            }
            if( x > i_emax )
            {
                i_emax = x;
            }
        }
        fGraph_AngularResolution68p[i_az]->GetPoint( 0, i_emin, y );
        fGraph_AngularResolution68p[i_az]->GetPoint( fGraph_AngularResolution68p[i_az]->GetN(), i_emax, y );
        for( int i = 0; i < Rec_nbins; i++ )
        {
            if( Rec_e0[i] > i_emin && Rec_e0[i] < i_emax )
            {
                Rec_angRes_p68[i] = fGraph_AngularResolution68p[i_az]->Eval( Rec_e0[i] );
            }
        }
    }
}

void VEffectiveAreaCalculator::setAngularResolutionGraph( unsigned int i_az, TGraphErrors* g, bool iAngContainment_80p )
{
    if( i_az < fGraph_AngularResolution68p.size() && g )
    {
        if( iAngContainment_80p )
        {
            fGraph_AngularResolution80p[i_az] = g;
        }
        else
        {
            fGraph_AngularResolution68p[i_az] = g;
        }
    }
}

void VEffectiveAreaCalculator::fillEcutSub( double iE, unsigned int iIndex )
{
    for( unsigned int i_az = 0; i_az < fVMinAz.size(); i_az++ )
    {
        for( unsigned int s = 0; s < fVSpectralIndex.size(); s++ )
        {
            if( s < hVEcutSub.size() && iIndex < hVEcutSub[s].size() && i_az < hVEcutSub[s][iIndex].size() )
            {
                hVEcutSub[s][iIndex][i_az]->Fill( iE, 1. );
            }
        }
    }
}


